#!/usr/bin/python3

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('XApp', '1.0')
from gi.repository import Gio, GLib, GObject, Gtk
import sys
import subprocess
import signal
import json
from pathlib import Path

signal.signal(signal.SIGINT, signal.SIG_DFL)


DBUS_NAME = "org.x.StatusIcon"
DBUS_PATH = "/org/x/StatusIcon"

class LayoutRow(GObject.Object):
    def __init__(self, shortname, json_info):
        super(LayoutRow, self).__init__()

        self.shortname = shortname
        self.name = json_info["name"]
        self.locale = json_info["locale"]

class OskLayoutTester():
    def __init__(self):
        self.window = None
        self.window = Gtk.Window()
        self.window.set_default_size(300, 400)
        self.window.connect("destroy", self.on_window_destroy)
        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                                margin=6,
                                spacing=0)

        # list stuff
        sw_frame = Gtk.Frame()
        sw = Gtk.ScrolledWindow(hadjustment=None, vadjustment=None)
        sw_frame.add(sw)

        self.store = Gio.ListStore(item_type=LayoutRow)

        self.list_box = Gtk.ListBox(activate_on_single_click=True)
        self.list_box.connect("row-activated", self.on_row_activated)
        self.list_box.bind_model(self.store, self.new_row_widget)

        sw.add(self.list_box)

        self.main_box.pack_start(sw_frame, True, True, 3)

        self.toggle_button = Gtk.Button(label="Toggle Keyboard")
        self.toggle_button.connect("clicked", self.toggle_keyboard)
        self.main_box.pack_start(self.toggle_button, False, False, 3)

        self.window.add(self.main_box)
        self.window.show_all()

        self.input_source_settings = Gio.Settings.new("org.cinnamon.desktop.input-sources")
        print("Saving original sources")
        self.orig_sources = self.input_source_settings.get_value("sources")

        self.shortname_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        self.load_layouts()

    def toggle_keyboard(self, button, data=None):
        subprocess.run(["cinnamon-dbus-command", "ToggleKeyboard"])

    def load_layouts(self):
        for file in Path.cwd().iterdir():
            if file.suffix != ".json":
                continue

            with open(file, "r") as f:
                json_info = json.load(f)
            self.store.append(LayoutRow(file.stem, json_info))

        self.store.sort(self.sort_rows)

    def sort_rows(self, a, b, data=None):
        if a.shortname < b.shortname:
            return -1
        elif a.shortname > b.shortname:
            return 1
        else:
            return 0

    def new_row_widget(self, item, data=None):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=2)

        label = Gtk.Label(label=f"<b>{item.shortname}</b>", visible=True, xalign=0, use_markup=True)
        self.shortname_size_group.add_widget(label)
        box.pack_start(label, False, False, 6)
        label = Gtk.Label(label=item.name, visible=True, xalign=0)
        box.pack_start(label, True, True, 0)
        label = Gtk.Label(label=f"<b>{item.locale}</b>", visible=True, xalign=0, use_markup=True)
        box.pack_end(label, False, False, 6)

        row = Gtk.ListBoxRow()
        row.add(box)
        row.name = item.shortname
        row.show_all()
        return row

    def on_row_activated(self, box, row, data=None):
        self.input_source_settings.set_value("sources", GLib.Variant("a(ss)", [("xkb", row.name)]))

    def on_window_destroy(self, widget, data=None):
        self.quit()

    def quit(self):
        if self.orig_sources is not None:
            print("Restoring original sources")
            self.input_source_settings.set_value("sources", self.orig_sources)

        Gtk.main_quit()

if __name__ == '__main__':
    test = OskLayoutTester()
    Gtk.main()



#!/usr/bin/python3

from gi.repository import Gtk
import pageutils

class ModulePage(pageutils.BaseListView):
    def __init__(self, parent):
        store = Gtk.ListStore(int, str, str, str)
        pageutils.BaseListView.__init__(self, store)
        self.parent = parent
        self.selected_path = None

        column = self.create_text_column(0, "ID")
        column.set_cell_data_func(self.renderer_text, self.cell_data_func_id)
        self.create_text_column(1, "Title")
        self.create_text_column(2, "WMClass")
        self.create_text_column(3, "Application")

        self.parent.lg_proxy.connect("signal::window-list-update", self.get_updates)
        self.parent.lg_proxy.connect("status-changed", self.on_status_change)

        self.tree_view.connect("row-activated", self.on_row_activated)
        self.tree_view.connect("button-press-event", self.on_button_press)

        # Popup menu
        self.popup = Gtk.Menu()

        inspect_window = Gtk.MenuItem('Inspect Window')
        label = inspect_window.get_children()[0]
        label.set_markup("<b>" + inspect_window.get_label() + "</b>")
        inspect_window.connect("activate", self.on_inspect_window)
        self.popup.append(inspect_window)

        self.inspect_app = Gtk.MenuItem('Inspect Application')
        self.inspect_app.connect("activate", self.on_inspect_application)
        self.popup.append(self.inspect_app)
        self.popup.show_all()

    def cell_data_func_id(self, column, cell, model, tree_iter, data=None):
        value = model.get_value(tree_iter, 0)
        cell.set_property("text", "w(%d) / a(%d)" %  (value, value))

    def on_row_activated(self, treeview, path, view_column):
        tree_iter = self.store.get_iter(path)
        obj_id = self.store.get_value(tree_iter, 0)
        title = self.store.get_value(tree_iter, 1)

        self.parent.pages["inspect"].inspect_element("w(%d)" % obj_id, "object", title, "<window>")

    def on_inspect_window(self, menu_item):
        tree_iter = self.store.get_iter(self.selected_path)
        obj_id = self.store.get_value(tree_iter, 0)
        title = self.store.get_value(tree_iter, 1)

        self.parent.pages["inspect"].inspect_element("w(%d)" % obj_id, "object", title, "<window>")

    def on_inspect_application(self, menu_item):
        tree_iter = self.store.get_iter(self.selected_path)
        obj_id = self.store.get_value(tree_iter, 0)
        application = self.store.get_value(tree_iter, 3)

        self.parent.pages["inspect"].inspect_element("a(%d)" % obj_id,
                                                     "object",
                                                     application,
                                                     "<application>")

    def on_button_press(self, treeview, event):
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                self.selected_path = path
                treeview.grab_focus()
                treeview.set_cursor(path, col, 0)

                tree_iter = self.store.get_iter(self.selected_path)
                app = self.store.get_value(tree_iter, 3)

                self.inspect_app.set_sensitive(app != "<untracked>")
                self.popup.popup(None, None, None, None, event.button, event.time)
            return True

    def on_status_change(self, proxy, online):
        if online:
            self.get_updates()

    def get_updates(self, proxy=None):
        self.store.clear()
        success, data = self.parent.lg_proxy.GetLatestWindowList()
        if success:
            try:
                for item in data:
                    self.store.append([int(item["id"]),
                                       item["title"],
                                       item["wmclass"],
                                       item["app"]])
            except Exception as exc:
                print(exc)



#!/usr/bin/python3

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("CDesktopEnums", "3.0")
from gi.repository import Gtk, Gdk, GLib, CDesktopEnums

from bin.SettingsWidgets import SidePage
from xapp.GSettingsWidgets import *


class Module:
    comment = _("Control mouse and touchpad settings")
    name = "mouse"
    category = "hardware"

    def __init__(self, content_box):
        keywords = _("mouse, touchpad, synaptic, double-click")
        sidePage = SidePage(_("Mouse and Touchpad"), "cs-mouse", keywords, content_box, module=self)
        self.sidePage = sidePage

    def on_module_selected(self):
        if not self.loaded:
            print("Loading Mouse module")

            self.sidePage.stack = SettingsStack()
            self.sidePage.add_widget(self.sidePage.stack)

            # Mouse

            page = SettingsPage()

            settings = page.add_section(_("General"))

            switch = GSettingsSwitch(_("Left handed (mouse buttons inverted)"), "org.cinnamon.desktop.peripherals.mouse", "left-handed")
            settings.add_row(switch)

            switch = GSettingsSwitch(_("Reverse scrolling direction"), "org.cinnamon.desktop.peripherals.mouse", "natural-scroll")
            settings.add_row(switch)

            switch = GSettingsSwitch(_("Show position of pointer when the Control key is pressed"), "org.cinnamon.desktop.peripherals.mouse", "locate-pointer")
            settings.add_row(switch)

            switch = GSettingsSwitch(_("Emulate middle click by clicking both left and right buttons"), "org.cinnamon.desktop.peripherals.mouse", "middle-click-emulation")
            settings.add_row(switch)

            switch = GSettingsSwitch(_("Paste the current selection when middle-click is pressed"), "org.cinnamon.desktop.interface", "gtk-enable-primary-paste")
            settings.add_row(switch)

            spin = GSettingsSpinButton(_("Drag-and-drop threshold"), "org.cinnamon.desktop.peripherals.mouse", "drag-threshold", _("pixels"), 1, 400)
            settings.add_row(spin)

            settings = page.add_section(_("Pointer size and speed"))

            widget = GSettingsRange(_("Size"), "org.cinnamon.desktop.interface", "cursor-size", _("Smaller"), _("Larger"), 5, 50, show_value=False)
            widget.add_mark(24.0, Gtk.PositionType.TOP, None)
            settings.add_row(widget)

            slider = GSettingsRange(_("Speed"), "org.cinnamon.desktop.peripherals.mouse", "speed", _("Slower"), _("Faster"), -1.0, 1.0, show_value=False)
            slider.content_widget.add_mark(0.0, Gtk.PositionType.TOP, None)

            settings.add_row(slider)

            accel_profiles = [
                ["default", _("Device default")],
                ["flat", _("Constant")],
                ["adaptive", _("Adaptive")]
            ]

            combo = GSettingsComboBox(_("Acceleration"), "org.cinnamon.desktop.peripherals.mouse", "accel-profile", accel_profiles, valtype=str)
            settings.add_row(combo)

            settings = page.add_section(_("Double-Click timeout"))

            slider = GSettingsRange(_("Timeout"), "org.cinnamon.desktop.peripherals.mouse", "double-click", _("Short"), _("Long"), 100, 1000, show_value=False)
            settings.add_row(slider)

            box = SettingsWidget()
            widget = Gtk.Button.new_with_label(_("Double-click test"))
            widget.connect("button-press-event", self.test_button_clicked)
            box.pack_start(widget, True, True, 0)
            settings.add_row(box)

            self.sidePage.stack.add_titled(page, "mouse", _("Mouse"))

            # Touchpad

            page = SettingsPage()

            send_events_modes = [
                ["enabled", _("Enabled")],
                ["disabled", _("Disabled")],
                ["disabled-on-external-mouse", _("Disabled when a mouse is attached")]
            ]

            combo = GSettingsComboBox(_("Touchpad is"), "org.cinnamon.desktop.peripherals.touchpad", "send-events", send_events_modes, valtype=str)
            combo.fill_row()
            page.pack_start(combo, False, True, 0)

            revealer = SettingsRevealer("org.cinnamon.desktop.peripherals.touchpad", "send-events", values=("enabled", "disabled-on-external-mouse"))
            page.pack_start(revealer, False, True, 0)

            settings = SettingsSection(_("General"))
            revealer.add(settings)

            switch = GSettingsSwitch(_("Tap to click"), "org.cinnamon.desktop.peripherals.touchpad", "tap-to-click")
            settings.add_row(switch)

            switch = GSettingsSwitch(_("Disable touchpad while typing"), "org.cinnamon.desktop.peripherals.touchpad", "disable-while-typing")
            settings.add_row(switch)

            clickpad_list = [
                ["default", _("Default device behavior")],
                ["none", _("Disabled")],
                ["areas", _("Virtual button areas along bottom of touchpad")],
                ["fingers", _("Use multiple fingers for right and middle click")]
            ]

            combo = GSettingsComboBox(_("Click actions"), "org.cinnamon.desktop.peripherals.touchpad", "click-method", clickpad_list, valtype=str)
            settings.add_row(combo)

            settings = SettingsSection(_("Scrolling"))
            revealer.add(settings)

            switch = GSettingsSwitch(_("Reverse scrolling direction"), "org.cinnamon.desktop.peripherals.touchpad", "natural-scroll")
            settings.add_row(switch)

            combo = ScrollMethodCombo()
            settings.add_row(combo)

            settings = SettingsSection(None)
            revealer.add(settings)

            slider = GSettingsRange(_("Speed"), "org.cinnamon.desktop.peripherals.touchpad", "speed", _("Slower"), _("Faster"), -1.0, 1.0, show_value=False)
            slider.content_widget.add_mark(0.0, Gtk.PositionType.TOP, None)
            settings.add_row(slider)

            self.sidePage.stack.add_titled(page, "touchpad", _("Touchpad"))

    def test_button_clicked(self, widget, event):
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            widget.set_label(_("Success!"))
            GLib.timeout_add(1000, self.reset_test_button, widget)
        return True

    def reset_test_button(self, widget):
        widget.set_label(_("Double-click test"))
        return False

class ScrollMethodCombo(ComboBox):
    def __init__(self):
        self.scrolling_list = [
            [0, _("Disabled")],
            [1, _("Two-finger scrolling")],
            [2, _("Edge scrolling")]
        ]

        super(ScrollMethodCombo, self).__init__(_("Scrolling method"), options=self.scrolling_list, valtype=int)

        self.touchpad_settings = Gio.Settings(schema_id="org.cinnamon.desktop.peripherals.touchpad")
        self.touchpad_settings.connect("changed::two-finger-scrolling-enabled", self.on_setting_changed)
        self.touchpad_settings.connect("changed::edge-scrolling-enabled", self.on_setting_changed)
        self.on_setting_changed(None, None)

        self.connect_widget_handlers()

    def on_my_value_changed(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            self.value = self.model[tree_iter][0]
            self.set_value(self.value)

    def set_value(self, value):
        if value == 0:
            self.touchpad_settings.set_boolean("two-finger-scrolling-enabled", False)
            self.touchpad_settings.set_boolean("edge-scrolling-enabled", False)
        elif value == 1:
            self.touchpad_settings.set_boolean("two-finger-scrolling-enabled", True)
            self.touchpad_settings.set_boolean("edge-scrolling-enabled", False)
        elif value == 2:
            self.touchpad_settings.set_boolean("two-finger-scrolling-enabled", False)
            self.touchpad_settings.set_boolean("edge-scrolling-enabled", True)

    def on_setting_changed(self, schema, key):
            finger = self.touchpad_settings.get_boolean("two-finger-scrolling-enabled")
            edge = self.touchpad_settings.get_boolean("edge-scrolling-enabled")

            if finger and edge:
                GLib.idle_add(self.touchpad_settings.set_boolean, ("edge-scrolling-enabled", False))
                # self.touchpad_settings.set_boolean("edge-scrolling-enabled", False)
                return
            elif finger:
                self.value = 1
            elif edge:
                self.value = 2
            else:
                self.value = 0

            try:
                self.content_widget.set_active_iter(self.option_map[self.value])
            except:
                self.content_widget.set_active_iter(None)



#!/usr/bin/python3

import subprocess
import gi
gi.require_version("Gtk", "3.0")

from gi.repository import Gtk
from bin.SettingsWidgets import SidePage, GSettingsDependencySwitch, DependencyCheckInstallButton, GSettingsSoundFileChooser
from xapp.GSettingsWidgets import *

DPI_FACTOR_LARGE         = 1.25
DPI_FACTOR_NORMAL        = 1.0

HIGH_CONTRAST_THEME      = "HighContrast"
KEY_TEXT_SCALING_FACTOR  = "text-scaling-factor"
KEY_GTK_THEME            = "gtk-theme"
KEY_GTK_THEME_BACKUP     = "gtk-theme-backup"
KEY_ICON_THEME           = "icon-theme"
KEY_ICON_THEME_BACKUP    = "icon-theme-backup"
KEY_WM_THEME             = "theme"
KEY_WM_THEME_BACKUP      = "theme-backup"


class Module:
    name = "accessibility"
    comment = _("Configure accessibility features")
    category = "prefs"

    def __init__(self, content_box):
        keywords = _("magnifier, talk, access, zoom, keys, contrast")
        sidePage = SidePage(_("Accessibility"), "cs-universal-access", keywords, content_box, module=self)
        self.sidePage = sidePage

    def on_module_selected(self):
        if not self.loaded:
            print("Loading Accessibility module")

            self.iface_settings = Gio.Settings(schema_id="org.cinnamon.desktop.interface")
            self.wm_settings = Gio.Settings(schema_id="org.cinnamon.desktop.wm.preferences")
            self.mag_settings = Gio.Settings(schema_id="org.cinnamon.desktop.a11y.magnifier")

            self.sidePage.stack = SettingsStack()
            self.sidePage.add_widget(self.sidePage.stack)

####    Visual

            page = SettingsPage()
            self.sidePage.stack.add_titled(page, "visual", _("Visual"))

# Visual Aids

            settings = page.add_section(_("Visual Aids"))

            switch = Switch(_("High contrast"))
            self.iface_settings.bind_with_mapping(KEY_GTK_THEME,
                                                  switch.content_widget, "active",
                                                  Gio.SettingsBindFlags.DEFAULT,
                                                  self.hi_con_get_mapping,
                                                  self.hi_con_set_mapping)
            settings.add_row(switch)

            switch = Switch(_("Large text"))
            self.iface_settings.bind_with_mapping(KEY_TEXT_SCALING_FACTOR,
                                                  switch.content_widget, "active",
                                                  Gio.SettingsBindFlags.DEFAULT,
                                                  self.lg_text_get_mapping,
                                                  self.lg_text_set_mapping)
            settings.add_row(switch)

            switch = GSettingsDependencySwitch(_("Screen reader"),
                                               "org.cinnamon.desktop.a11y.applications",
                                               "screen-reader-enabled",
                                               None,
                                               ["orca"],
                                               ["gnome-orca"])
            settings.add_row(switch)

            def config_orca(button):
                subprocess.Popen(['orca', '--setup'])

            button = Button(_("Configure screen reader"), config_orca)
            settings.add_reveal_row(button, "org.cinnamon.desktop.a11y.applications", "screen-reader-enabled")

# Desktop Zoom

            settings = page.add_section(_("Desktop Zoom"))

            switch = GSettingsSwitch(_("Enable zoom"), "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")
            settings.add_row(switch)

            spin = GSettingsSpinButton(_("Magnification"), "org.cinnamon.desktop.a11y.magnifier", "mag-factor", None, 1.0, 15.0, step=0.5)
            settings.add_reveal_row(spin, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            zoom_key_options = [["", _("Disabled")], ["<Alt>", "<Alt>"],["<Super>", "<Super>"],["<Control>", "<Control>"], ["<Shift>", "<Shift>"]]
            widget = GSettingsComboBox(_("Mouse wheel modifier"), "org.cinnamon.desktop.wm.preferences", "mouse-button-zoom-modifier", zoom_key_options)
            widget.set_tooltip_text(_("While this modifier is pressed, mouse scrolling will increase or decrease zoom."))
            settings.add_reveal_row(widget, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            switch = GSettingsSwitch(_("Scroll at screen edges"), "org.cinnamon.desktop.a11y.magnifier", "scroll-at-edges")
            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            mouse_track_options = [["centered",     _("Keep cursor centered")],
                                   ["proportional", _("Cursor moves with contents")],
                                   ["push",         _("Cursor pushes contents around")]]

            widget = GSettingsComboBox(_("Mouse tracking mode"), "org.cinnamon.desktop.a11y.magnifier", "mouse-tracking", mouse_track_options)
            settings.add_reveal_row(widget, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            switch = GSettingsSwitch(_("Lens mode"), "org.cinnamon.desktop.a11y.magnifier", "lens-mode")
            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            self.zoom_stack = SettingsStack()
            self.zoom_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

            lens_shape_options = [["square",          _("Square")],
                                  ["horizontal",      _("Horizontal strip")],
                                  ["vertical",        _("Vertical strip")]]

            widget = GSettingsComboBox(_("Lens shape"), "org.cinnamon.desktop.a11y.magnifier", "lens-shape", lens_shape_options)
            self.zoom_stack.add_named(widget, "shape")

            screen_pos_options = [["full-screen",     _("Full screen")],
                                  ["top-half",        _("Top half")],
                                  ["bottom-half",     _("Bottom half")],
                                  ["left-half",       _("Left half")],
                                  ["right-half",      _("Right half")]]

            widget = GSettingsComboBox(_("Screen position"), "org.cinnamon.desktop.a11y.magnifier", "screen-position", screen_pos_options)
            self.zoom_stack.add_named(widget, "screen")

            settings.add_reveal_row(self.zoom_stack, "org.cinnamon.desktop.a11y.applications", "screen-magnifier-enabled")

            self.mag_settings.bind_with_mapping("lens-mode",
                                                self.zoom_stack, "visible-child-name",
                                                Gio.SettingsBindFlags.GET,
                                                self.zoom_stack_get,
                                                None)

            if self.mag_settings.get_boolean("lens-mode"):
                self.zoom_stack.set_visible_child_name("shape")
            else:
                self.zoom_stack.set_visible_child_name("screen")

#### Keyboard

            page = SettingsPage()
            self.sidePage.stack.add_titled(page, "keyboard", _("Keyboard"))

# Virtual keyboard

            settings = page.add_section(_("Virtual keyboard"))

            switch = GSettingsSwitch(_("Enable the on-screen keyboard"),
                                     "org.cinnamon.desktop.a11y.applications",
                                     "screen-keyboard-enabled",
                                     None)

            settings.add_row(switch)

            keyboard_position_options = [["bottom",     _("At the bottom of the screen")],
                                         ["top",     _("At the top of the screen")]]
            widget = GSettingsComboBox(_("Keyboard position"), "org.cinnamon.keyboard", "keyboard-position", keyboard_position_options)
            settings.add_reveal_row(widget, "org.cinnamon.desktop.a11y.applications", "screen-keyboard-enabled")

            keyboard_size_options = [[4,     _("A quarter of the screen")],
                                     [3,     _("A third of the screen")]]

            widget = GSettingsComboBox(_("Keyboard size"), "org.cinnamon.keyboard", "keyboard-size", keyboard_size_options)
            settings.add_reveal_row(widget, "org.cinnamon.desktop.a11y.applications", "screen-keyboard-enabled")

            activation_mode_options = [["accessible", _("Show the keyboard any time something expects input")],
                                       ["on-demand",  _("Show keyboard only when the user activates it")]]

            widget = GSettingsComboBox(_("Activation mode"), "org.cinnamon.keyboard", "activation-mode", activation_mode_options)
            settings.add_reveal_row(widget, "org.cinnamon.desktop.a11y.applications", "screen-keyboard-enabled")

# Keyboard indicators

            settings = page.add_section(_("Keyboard indicators"))

            switch = GSettingsSwitch(_("Use visual indicator on Caps and Num Lock"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "togglekeys-enable-osd")

            settings.add_row(switch)

            switch = GSettingsSwitch(_("Use audio indicator on Caps and Num Lock"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "togglekeys-enable-beep")

            settings.add_row(switch)

            sound_picker = GSettingsSoundFileChooser(_("Sound to use Caps or Num Lock on"),
                                                     "org.cinnamon.desktop.a11y.keyboard",
                                                     "togglekeys-sound-on")
            settings.add_reveal_row(sound_picker, "org.cinnamon.desktop.a11y.keyboard", "togglekeys-enable-beep")

            sound_picker = GSettingsSoundFileChooser(_("Sound to use Caps or Num Lock off"),
                                                     "org.cinnamon.desktop.a11y.keyboard",
                                                     "togglekeys-sound-off")
            settings.add_reveal_row(sound_picker, "org.cinnamon.desktop.a11y.keyboard", "togglekeys-enable-beep")

            settings = page.add_section(_("Event feedback (required for typing assistance alerts)"))

            switch = GSettingsSwitch(_("Enable visual alerts"),
                                     "org.cinnamon.desktop.wm.preferences",
                                     "visual-bell")
            settings.add_row(switch)

            visual_bell_options = [["fullscreen-flash",     _("Flash the entire monitor")],
                                   ["frame-flash", _("Flash the active window")]]

            widget = GSettingsComboBox(_("Visual style"), "org.cinnamon.desktop.wm.preferences", "visual-bell-type", visual_bell_options)
            settings.add_reveal_row(widget, "org.cinnamon.desktop.wm.preferences", "visual-bell")

            switch = GSettingsSwitch(_("Enable audio alerts"),
                                     "org.cinnamon.desktop.wm.preferences",
                                     "audible-bell")
            settings.add_row(switch)

            sound_picker = GSettingsSoundFileChooser(_("Sound to use for window alerts"),
                                                     "org.cinnamon.desktop.wm.preferences",
                                                     "bell-sound")
            settings.add_reveal_row(sound_picker, "org.cinnamon.desktop.wm.preferences", "audible-bell")

####    Typing Assistance

            page = SettingsPage()
            self.sidePage.stack.add_titled(page, "typing", _("Typing assistance"))

# Stickykeys

            settings = page.add_section(_("Sticky keys"))

            switch = GSettingsSwitch(_("Treat a sequence of modifier keys as a combination"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "stickykeys-enable",
                                     None)

            settings.add_row(switch)

            switch = GSettingsSwitch(_("Disable if two modifiers are pressed together"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "stickykeys-two-key-off",
                                     None)

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "stickykeys-enable")

            switch = GSettingsSwitch(_("Alert when a modifier key is pressed"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "stickykeys-modifier-beep",
                                     "org.cinnamon.desktop.sound/event-sounds")

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "stickykeys-enable")

# Slowkeys

            settings = page.add_section(_("Slow keys"))

            switch = GSettingsSwitch(_("Put a delay between when a key is pressed and when it is accepted"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "slowkeys-enable",
                                     None)

            settings.add_row(switch)

            switch = GSettingsSwitch(_("Alert when a key is pressed"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "slowkeys-beep-press",
                                     "org.cinnamon.desktop.sound/event-sounds")

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "slowkeys-enable")

            switch = GSettingsSwitch(_("Alert when a key is accepted"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "slowkeys-beep-accept",
                                     "org.cinnamon.desktop.sound/event-sounds")

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "slowkeys-enable")

            switch = GSettingsSwitch(_("Alert when a key is rejected"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "slowkeys-beep-reject",
                                     "org.cinnamon.desktop.sound/event-sounds")

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "slowkeys-enable")

            slider = GSettingsRange(_("Acceptance delay"),
                                    "org.cinnamon.desktop.a11y.keyboard",
                                    "slowkeys-delay",
                                    _("Short"), _("Long"),
                                    0, 500, 10, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.keyboard", "slowkeys-enable")

# Bouncekeys

            settings = page.add_section(_("Bounce keys"))

            switch = GSettingsSwitch(_("Ignore fast duplicate keypresses"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "bouncekeys-enable",
                                     None)

            settings.add_row(switch)

            switch = GSettingsSwitch(_("Alert when a key is rejected"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "bouncekeys-beep-reject",
                                     "org.cinnamon.desktop.sound/event-sounds")

            settings.add_reveal_row(switch, "org.cinnamon.desktop.a11y.keyboard", "bouncekeys-enable")

            slider = GSettingsRange(_("Acceptance delay"),
                                    "org.cinnamon.desktop.a11y.keyboard",
                                    "bouncekeys-delay",
                                    _("Short"), _("Long"),
                                    0, 900, 10, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.keyboard", "bouncekeys-enable")

####    Pointing and Clicking

            page = SettingsPage()
            self.sidePage.stack.add_titled(page, "mouse", _("Mouse"))

# Mousekeys

            settings = page.add_section(_("Mouse keys"))

            switch = GSettingsSwitch(_("Control the pointer using the keypad"),
                                     "org.cinnamon.desktop.a11y.keyboard",
                                     "mousekeys-enable",
                                     None)

            settings.add_row(switch)

            slider = GSettingsRange(_("Initial delay"),
                                    "org.cinnamon.desktop.a11y.keyboard",
                                    "mousekeys-init-delay",
                                    _("Shorter"), _("Longer"),
                                    10, 2000, 10, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.keyboard", "mousekeys-enable")

            slider = GSettingsRange(_("Acceleration time"),
                                    "org.cinnamon.desktop.a11y.keyboard",
                                    "mousekeys-accel-time",
                                    _("Shorter"), _("Longer"),
                                    10, 2000, 10, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.keyboard", "mousekeys-enable")

            slider = GSettingsRange(_("Maximum speed"),
                                    "org.cinnamon.desktop.a11y.keyboard",
                                    "mousekeys-max-speed",
                                    _("Slower"), _("Faster"),
                                    1, 500, 1, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.keyboard", "mousekeys-enable")

# Secondary click

            settings = page.add_section(_("Simulated secondary click"))

            switch = GSettingsSwitch(_("Trigger a secondary click by holding down the primary button"),
                                     "org.cinnamon.desktop.a11y.mouse",
                                     "secondary-click-enabled",
                                     None)

            settings.add_row(switch)

            slider = GSettingsRange(_("Acceptance delay"),
                                    "org.cinnamon.desktop.a11y.mouse",
                                    "secondary-click-time",
                                    _("Shorter"), _("Longer"),
                                    0.5, 3.0, 0.1, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.mouse", "secondary-click-enabled")

# Hover Click

            settings = page.add_section(_("Hover click"))

            switch = GSettingsSwitch(_("Trigger a click when the pointer hovers"),
                                     "org.cinnamon.desktop.a11y.mouse",
                                     "dwell-click-enabled",
                                     None)

            settings.add_row(switch)

            slider = GSettingsRange(_("Delay"),
                                    "org.cinnamon.desktop.a11y.mouse",
                                    "dwell-time",
                                    _("Short"), _("Long"),
                                    0.2, 3.0, 0.1, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.mouse", "dwell-click-enabled")

            slider = GSettingsRange(_("Motion threshold"),
                                    "org.cinnamon.desktop.a11y.mouse",
                                    "dwell-threshold",
                                    _("Small"), _("Large"),
                                    1, 30, 1, show_value=False)

            settings.add_reveal_row(slider, "org.cinnamon.desktop.a11y.mouse", "dwell-click-enabled")

    def zoom_stack_get(self, lens_mode):
        ret = "screen"

        if lens_mode:
            ret = "shape"
        else:
            ret = "screen"

        return ret

    def hi_con_get_mapping(self, string):
        return string == HIGH_CONTRAST_THEME

    def hi_con_set_mapping(self, active):
        ret = None

        if active:
            ret = HIGH_CONTRAST_THEME

            theme = self.iface_settings.get_string(KEY_GTK_THEME)
            self.iface_settings.set_string(KEY_GTK_THEME_BACKUP, theme)

            theme = self.iface_settings.get_string(KEY_ICON_THEME)
            self.iface_settings.set_string(KEY_ICON_THEME_BACKUP, theme)
            self.iface_settings.set_string(KEY_ICON_THEME, HIGH_CONTRAST_THEME)

            theme = self.wm_settings.get_string(KEY_WM_THEME)
            self.wm_settings.set_string(KEY_WM_THEME_BACKUP, theme)
            self.wm_settings.set_string(KEY_WM_THEME, HIGH_CONTRAST_THEME)
        else:
            ret = self.iface_settings.get_string(KEY_GTK_THEME_BACKUP)

            theme = self.iface_settings.get_string(KEY_ICON_THEME_BACKUP)
            self.iface_settings.set_string(KEY_ICON_THEME, theme)

            theme = self.wm_settings.get_string(KEY_WM_THEME_BACKUP)
            self.wm_settings.set_string(KEY_WM_THEME, theme)

        return ret

    def lg_text_get_mapping(self, factor):
        return factor > DPI_FACTOR_NORMAL

    def lg_text_set_mapping(self, active):
        ret = None

        if active:
            ret = DPI_FACTOR_LARGE
        else:
            ret = self.iface_settings.get_default_value(KEY_TEXT_SCALING_FACTOR).get_double()

        return ret



#!/usr/bin/python3

import os
import sys

try:
    from gi.repository import Gio, Gtk, GObject, Gdk, GdkPixbuf, GLib
    import html
    import subprocess
    import threading
except Exception as error_message:
    print(error_message)
    sys.exit(1)

from cinnamon.harvester import Harvester, AbortedError, URL_SPICES_HOME

ABORT_NONE = 0
ABORT_ERROR = 1
ABORT_USER = 2

def ui_thread_do(callback, *args):
    GLib.idle_add(callback, *args, priority=GLib.PRIORITY_DEFAULT)

class Spice_Harvester(GObject.Object):
    __gsignals__ = {
        'installed-changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'status-changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'cache-loaded': (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self, collection_type, window=None):
        super().__init__()
        self.collection_type = collection_type
        self.window = window

        self.themes = collection_type == 'theme'
        self.actions = collection_type == 'action'

        self._h = Harvester(collection_type)

        self.updates_available = []

        self._proxy = None
        self._proxy_deferred_actions = []
        self.running_uuids = []

        self.jobs = []
        self.progressbars = []
        self.processing_jobs = False
        self.current_job = None
        self.total_jobs = 0

        self.abort_download = ABORT_NONE

        self.settings = self._h.settings
        self.enabled_key = self._h.enabled_key

        if not self.themes:
            self.settings.connect(f'changed::{self.enabled_key}', self._update_status)

        self._update_status()
        self._generate_update_list()

        self.monitorId = 0
        self.monitor = None
        try:
            self.monitor = Gio.File.new_for_path(self._h.install_folder).monitor_directory(0, None)
            self.monitorId = self.monitor.connect('changed', self._directory_changed)
        except Exception as e:
            # File monitors can fail when the OS runs out of file handles
            print(e)

        # Nemo discovers actions via filesystem monitoring and reloads them on its own.
        if not self.actions:
            try:
                Gio.DBusProxy.new_for_bus(Gio.BusType.SESSION, Gio.DBusProxyFlags.NONE, None,
                                          'org.Cinnamon', '/org/Cinnamon', 'org.Cinnamon',
                                          None, self._on_proxy_ready, None)
            except GLib.Error as e:
                print(e)

    def _on_proxy_ready(self, obj, result, data=None):
        try:
            self._proxy = Gio.DBusProxy.new_for_bus_finish(result)
            self._proxy.connect('g-signal', self._on_signal)

            if self._proxy.get_name_owner():
                self.send_deferred_proxy_calls()
            else:
                print("org.Cinnamon proxy created, but no owner - is Cinnamon running?")
        except GLib.Error as e:
            print(f"Could not establish proxy for org.Cinnamon: {e}")

        self._update_status()

    def _on_signal(self, proxy, sender_name, signal_name, params):
        if signal_name == "RunStateChanged":
            if self._proxy.GetRunState() == 2:
                self.send_deferred_proxy_calls()
        elif signal_name in ("XletsLoadedComplete", "XletAddedComplete"):
            self._update_status()

    def send_deferred_proxy_calls(self):
        if self._proxy.get_name_owner() and self._proxy_deferred_actions:
            for command, args in self._proxy_deferred_actions:
                getattr(self._proxy, command)(*args)

            self._proxy_deferred_actions = []

    def send_proxy_signal(self, command, *args):
        """ sends a command over dbus"""
        if self._proxy is None or not self._proxy.get_name_owner():
            self._proxy_deferred_actions.append((command, args))
        else:
            getattr(self._proxy, command)(*args)

    def _update_status(self, *args):
        try:
            if self._proxy and self._proxy.get_name_owner() and not self.actions:
                self.running_uuids = self._proxy.GetRunningXletUUIDs('(s)', self.collection_type)
            else:
                self.running_uuids = []
        except:
            self.running_uuids = []
        self.emit('status-changed')

    def open_spice_page(self, uuid):
        """ opens to the web page of the given uuid"""
        spice_id = self._h.index_cache[uuid]['spices-id']
        subprocess.run(['/usr/bin/xdg-open',
                       f"{URL_SPICES_HOME}/{self.collection_type}s/view/{spice_id}"],
                       check=True)

    def get_progressbar(self):
        """ returns a Gtk.Widget that can be added to the application. This widget will show the
            progress of any asynchronous actions taking place (ie. refreshing the cache or
            downloading an applet)"""
        progressbar = Gtk.ProgressBar()
        progressbar.set_show_text(True)
        progressbar.set_text('')
        progressbar.set_fraction(0)

        revealer = Gtk.Revealer()
        revealer.add(progressbar)
        revealer.set_transition_type(Gtk.RevealerTransitionType.SLIDE_DOWN)
        revealer.set_transition_duration(150)
        progressbar.revealer = revealer

        self.progressbars.append(progressbar)

        return revealer

    def _set_progressbar_text(self, text):
        for progressbar in self.progressbars:
            progressbar.set_text(text)

    def _set_progressbar_fraction(self, fraction):
        for progressbar in self.progressbars:
            progressbar.set_fraction(fraction)
        if self.window:
            self.window.set_progress(int(fraction * 100))

    def _set_progressbar_visible(self, visible):
        for progressbar in self.progressbars:
            progressbar.revealer.set_reveal_child(visible)

    def _update_progress(self, count, block_size, total_size):
        """ Progress callback handed to harvester. Runs on a worker thread.
            Raise AbortedError to halt an in-progress download."""
        if self._is_aborted():
            raise AbortedError()

        if block_size == 1:
            # Per-item progress (thumbnail downloads): count and total_size are item counts.
            fraction = count / total_size if total_size else 0
            text = _("Downloading images:") + f" {count}/{total_size}"
            ui_thread_do(self._set_progressbar_text, text)
        else:
            # Byte-level progress (zip download): count is chunks, block_size is chunk size.
            if total_size > 0:
                fraction = (count * block_size) / total_size
                if fraction > 1.0:
                    fraction = 1.0
            else:
                fraction = 0

        ui_thread_do(self._set_progressbar_fraction, fraction)

    # Jobs are added by calling _push_job. _process_job and _advance_queue
    # form a wrapper that runs the job in its own thread.
    def _push_job(self, job):
        self.total_jobs += 1
        job['job_number'] = self.total_jobs
        self.jobs.append(job)
        if not self.processing_jobs:
            self._advance_queue()

    def _process_job(self, job):
        job['result'] = job['func'](job)
        if job.get('callback'):
            GLib.idle_add(job['callback'], job)
        GLib.idle_add(self._advance_queue)

    def _advance_queue(self):
        if self.monitorId > 0:
            self.monitor.disconnect(self.monitorId)
            self.monitorId = 0

        self.processing_jobs = True

        self._set_progressbar_fraction(0)

        if len(self.jobs) > 0:
            # A new job is starting; clear any abort flag left over from a
            # previous job so its progress callback doesn't immediately fire
            # AbortedError.
            self.abort_download = ABORT_NONE
            self._set_progressbar_visible(True)
            job = self.jobs.pop(0)
            self.current_job = job
            text = job['progress_text']
            if self.total_jobs > 1:
                text += f" ({job['job_number']}/{self.total_jobs})"
            self._set_progressbar_text(text)
            job_thread = threading.Thread(target=self._process_job, args=(job,))
            job_thread.start()
        else:
            self.processing_jobs = False
            self.current_job = None
            self.total_jobs = 0
            self.abort_download = ABORT_NONE
            self._set_progressbar_visible(False)
            self._set_progressbar_text('')
            if self.monitor is not None:
                try:
                    self.monitorId = self.monitor.connect('changed', self._directory_changed)
                except Exception as e:
                    # File monitors can fail when the OS runs out of file handles
                    print(e)

            self.updates_available = [u.uuid for u in self._h.get_updates()]
            self.emit("installed-changed")

    def _directory_changed(self, *args):
        # External filesystem change — meta_map may be stale, full reload.
        self.updates_available = [u.uuid for u in self._h.reload()]
        self.emit("installed-changed")

    def get_installed(self):
        """ returns a dictionary of the metadata by uuid of all installed spices"""
        return self._h.meta_map

    def get_is_installed(self, uuid):
        """ returns a boolean specifying whether the given spice is installed or not"""
        return uuid in self._h.meta_map

    def get_has_update(self, uuid):
        """ returns a boolean indicating whether the given spice has an update available"""
        return self._h.has_update(uuid)

    def get_enabled(self, uuid):
        """ returns the number of instances currently enabled"""
        return self._h.get_enabled(uuid)

    def get_is_running(self, uuid):
        """ checks whether the spice is currently running (it may be enabled but not running if
            there was an error in initialization)"""
        return uuid in self.running_uuids

    def get_n_updates(self):
        """ returns the number of available updates"""
        return len(self.updates_available)

    def get_cache(self):
        """ retrieves a copy of the index cache """
        return self._h.index_cache

    def _generate_update_list(self):
        self.updates_available = [u.uuid for u in self._h.get_updates()]

    def refresh_cache(self):
        """ downloads an updated version of the index and assets"""
        job = {'func': self._do_refresh_cache, 'callback': self._on_refresh_cache_done}
        job['progress_text'] = _("Refreshing the cache")
        self._push_job(job)

    def _do_refresh_cache(self, job):
        try:
            self._h.refresh(full=True, force=True, progress_callback=self._update_progress)
        except AbortedError:
            return None
        except ValueError as e:
            if not self.abort_download:
                self.errorMessage(_("Something went wrong with the spices download. Please try refreshing the list again."), str(e))
                self.abort(ABORT_ERROR)
            return None
        except Exception as e:
            if not self.abort_download:
                self.errorMessage(_("An error occurred while trying to access the server. Please try again in a little while."), str(e))
                self.abort(ABORT_ERROR)
            return None
        return True

    def _on_refresh_cache_done(self, job):
        self._generate_update_list()
        self.emit('cache-loaded')

    def install(self, uuid):
        """ downloads and installs the given extension"""
        _callback = None if self.actions else self._install_finished
        job = {'uuid': uuid, 'func': self._install, 'callback': _callback}
        job['progress_text'] = _("Installing %s") % uuid
        self._push_job(job)

    def _install(self, job):
        uuid = job['uuid']
        try:
            self._h.install(uuid, progress_callback=self._update_progress)
        except AbortedError:
            return None
        except Exception as e:
            if not self.abort_download:
                self.errorMessage(_("An error occurred during the installation of %s. Please report this incident to its developer.") % uuid, str(e))
                self.abort(ABORT_ERROR)
            return None

        return True

    def _install_finished(self, job):
        if job.get('result') is None:
            return
        uuid = job['uuid']
        if self.get_enabled(uuid):
            if self._proxy:
                self._proxy.ReloadXlet('(ss)', uuid, self.collection_type.upper())
            else:
                self.send_proxy_signal('ReloadXlet', '(ss)', uuid, self.collection_type.upper())

    def uninstall(self, uuid):
        """ uninstalls and removes the given extension"""
        job = {'uuid': uuid, 'func': self._uninstall}
        job['progress_text'] = _("Uninstalling %s") % uuid
        self._push_job(job)

    def _uninstall(self, job):
        try:
            self._h.uninstall(job['uuid'])
        except Exception as error:
            self.errorMessage(_("A problem occurred while removing %s.") % job['uuid'], str(error))

    def update_all(self):
        """ applies all available updates"""
        for uuid in list(self.updates_available):
            self.install(uuid)

    def abort(self, abort_type=ABORT_USER):
        """ trigger in-progress download to halt"""
        self.abort_download = abort_type
        # Drop any queued jobs; the in-flight one (if any) will exit on its next
        # progress callback by raising AbortedError.
        self.jobs = []

    def _is_aborted(self):
        return self.abort_download != ABORT_NONE

    def _ui_error_message(self, msg, detail=None):
        dialog = Gtk.MessageDialog(transient_for=self.window,
                                   modal=True,
                                   message_type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK)
        markup = msg
        if detail is not None:
            markup += _("\n\nDetails:  %s") % (str(detail))
        esc = html.escape(markup)
        dialog.set_markup(esc)
        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def errorMessage(self, msg, detail=None):
        ui_thread_do(self._ui_error_message, msg, detail)

    def enable_extension(self, uuid, panel=1, box='right', position=0):
        if self.collection_type == 'desklet':
            screen = Gdk.Screen.get_default()
            primary = screen.get_primary_monitor()
            primary_rect = screen.get_monitor_geometry(primary)
            self._h.enable(uuid, desklet_x=primary_rect.x + 100, desklet_y=primary_rect.y + 100)
        else:
            self._h.enable(uuid, panel=panel, box=box, position=position)

    def disable_extension(self, uuid):
        self._h.disable(uuid)

    def get_icon(self, uuid):
        """ gets the icon for a given uuid"""
        try:
            if self.themes:
                file_path = os.path.join(self._h.cache_folder, os.path.basename(self._h.index_cache[uuid]['screenshot']))
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(file_path, 100, -1, True)
            else:
                file_path = os.path.join(self._h.cache_folder, os.path.basename(self._h.index_cache[uuid]['icon']))
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(file_path, 24, 24, True)

            return Gtk.Image.new_from_pixbuf(pixbuf)
        except Exception:
            print("There was an error processing one of the images. Try refreshing the cache.")
            return Gtk.Image.new_from_icon_name('image-missing', 2)



from gi.repository import GObject

from backends.cinnamon import CinnamonBackend
from backends.x11 import X11Backend


class Backend(GObject.Object):
    """Routes screenshot calls to the cinnamon DBus backend when it is
    available, falling back to the X11 backend for full-screen captures
    otherwise. Emits 'online-changed' whenever the cinnamon service appears
    or disappears so the UI can adjust which modes are offered."""

    __gsignals__ = {
        'online-changed': (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self):
        super().__init__()
        self._cinnamon = CinnamonBackend()
        self._x11 = X11Backend()
        self._cinnamon.connect('online-changed', self._on_cinnamon_online_changed)

    def _on_cinnamon_online_changed(self, _backend, online):
        self.emit('online-changed', online)

    def is_available(self):
        return self._cinnamon.is_available()

    def _active(self):
        return self._cinnamon if self._cinnamon.is_available() else self._x11

    def screenshot(self, include_pointer, on_done):
        self._active().screenshot(include_pointer, on_done)

    def screenshot_window(self, include_pointer, include_shadow, on_done):
        self._active().screenshot_window(include_pointer, include_shadow, on_done)

    def screenshot_window_by_id(self, window_id, include_pointer, include_shadow, on_done):
        self._active().screenshot_window_by_id(window_id, include_pointer, include_shadow, on_done)

    def screenshot_area(self, x, y, w, h, include_pointer, on_done):
        self._active().screenshot_area(x, y, w, h, include_pointer, on_done)

    def flash_area(self, x, y, w, h):
        self._active().flash_area(x, y, w, h)

    def select_area(self, on_done):
        self._active().select_area(on_done)

    def select_window(self, on_done):
        self._active().select_window(on_done)



import os

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, GLib, Gtk

SCHEMA_ID = 'org.cinnamon.screenshot'

DELAY_KEY = 'delay'
INCLUDE_POINTER_KEY = 'include-pointer'
INCLUDE_SHADOW_KEY = 'include-shadow'
SAVE_DIRECTORY_KEY = 'save-directory'
DEFAULT_FILE_TYPE_KEY = 'default-file-type'
LAUNCH_FILE_MANAGER_KEY = 'launch-file-manager-after-save'
AUTOSAVE_TO_FILE_KEY = 'autosave-to-file'
AUTOSAVE_TO_CLIPBOARD_KEY = 'autosave-to-clipboard'

settings = Gio.Settings.new(SCHEMA_ID)
if not settings.get_string(SAVE_DIRECTORY_KEY):
    path = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES) or GLib.get_home_dir()
    if path:
        settings.set_string(SAVE_DIRECTORY_KEY, Gio.File.new_for_path(path).get_uri())

def get_delay():
    return settings.get_int(DELAY_KEY)

def set_delay(value):
    settings.set_int(DELAY_KEY, value)

def get_include_pointer():
    return settings.get_boolean(INCLUDE_POINTER_KEY)

def set_include_pointer(value):
    settings.set_boolean(INCLUDE_POINTER_KEY, value)

def get_include_shadow():
    return settings.get_boolean(INCLUDE_SHADOW_KEY)

def get_save_directory_uri():
    return settings.get_string(SAVE_DIRECTORY_KEY)

def get_save_directory():
    """Always returns an existing directory. Tries the user's saved
    preference first; falls back to home if that path is missing or invalid."""
    uri = get_save_directory_uri()
    if uri:
        try:
            path = Gio.File.new_for_uri(uri).get_path() or ''
        except Exception:
            path = ''
        if path and os.path.isdir(path):
            return path
    return GLib.get_home_dir()

def get_default_file_type():
    return settings.get_string(DEFAULT_FILE_TYPE_KEY) or 'png'

def set_default_file_type(value):
    settings.set_string(DEFAULT_FILE_TYPE_KEY, value)

def get_launch_file_manager():
    return settings.get_boolean(LAUNCH_FILE_MANAGER_KEY)

def get_autosave_to_file():
    return settings.get_boolean(AUTOSAVE_TO_FILE_KEY)

def get_autosave_to_clipboard():
    return settings.get_boolean(AUTOSAVE_TO_CLIPBOARD_KEY)

_current_window = None


def open_preferences(parent):
    global _current_window
    if _current_window is not None:
        _current_window.window.present()
        return
    _current_window = PreferencesWindow(parent)


class PreferencesWindow:
    def __init__(self, parent):
        from xapp.GSettingsWidgets import GSettingsComboBox, GSettingsFileChooser, GSettingsSwitch
        from xapp.SettingsWidgets import SettingsPage

        self.window = Gtk.Window(title=_('Preferences'))
        self.window.set_transient_for(parent)
        self.window.set_destroy_with_parent(True)
        self.window.set_default_size(600, 200)
        self.window.connect('destroy', self._on_destroy)

        page = SettingsPage()
        self.window.add(page)

        section = page.add_section(_('Saving'))

        size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)
        section.add_row(GSettingsFileChooser(
            _('Default save folder'),
            SCHEMA_ID, SAVE_DIRECTORY_KEY,
            size_group=size_group, dir_select=True,
        ))
        section.add_row(GSettingsComboBox(
            _('Default file type'),
            SCHEMA_ID, DEFAULT_FILE_TYPE_KEY,
            options=[
                ('png',  'PNG'),
                ('jpg',  'JPEG'),
                ('bmp',  'BMP'),
                ('tiff', 'TIFF'),
            ],
            valtype=str,
            size_group=size_group,
        ))
        section.add_row(GSettingsSwitch(
            _('Launch file manager after saving'),
            SCHEMA_ID, LAUNCH_FILE_MANAGER_KEY,
        ))

        section_behavior = page.add_section(_('Behavior'))
        section_behavior.add_row(GSettingsSwitch(
            _('Autosave to default folder'),
            SCHEMA_ID, AUTOSAVE_TO_FILE_KEY,
        ))
        section_behavior.add_row(GSettingsSwitch(
            _('Copy to clipboard when saving'),
            SCHEMA_ID, AUTOSAVE_TO_CLIPBOARD_KEY,
        ))

        self.window.show_all()

    def _on_destroy(self, _w):
        global _current_window
        _current_window = None



#!/usr/bin/python3

import os
import subprocess

link_path = os.path.join('etc', 'xdg', 'menus', 'cinnamon-applications-merged')
dest = os.environ.get('DESTDIR')

if dest:
    link_path = os.path.join(dest, link_path)
else:
    link_path = os.path.join('/', link_path)

if os.path.lexists(link_path):
    print('%s already exists, skipping symlink creation' % link_path)
else:
    print('adding symlink %s...' % link_path)
    subprocess.call(['ln', '-s', 'applications-merged', link_path])



#!/usr/bin/python3

from pathlib import Path
import sys

from bin.ExtensionCore import ManageSpicesPage, DownloadSpicesPage
from bin.Spices import Spice_Harvester
from bin.SettingsWidgets import SidePage
from xapp.GSettingsWidgets import *
from gi.repository import GLib

class Module:
    comment = _("Manage your actions")
    name = "actions"
    category = "prefs"

    def __init__(self, content_box):
        self.window = None
        self.sidePage = ActionsViewSidePage(content_box, self)

    def on_module_selected(self):
        if not self.loaded:
            print("Loading Actions module")
            self.sidePage.load(self.window)

    def _setParentRef(self, window):
        self.window = window


class ActionsViewSidePage(SidePage):
    collection_type = "action"

    def __init__(self, content_box, module):
        self.RemoveString = ""
        keywords = _("action")

        super().__init__(_("Actions"), "cs-actions", keywords,
                         content_box, module=module)

    def load(self, window):
        self.window = window

        self.spices = Spice_Harvester(self.collection_type, self.window)

        self.stack = SettingsStack()
        self.add_widget(self.stack)
        self.stack.expand = True

        manage_extensions_page = ManageActionsPage(self, self.spices, self.window)
        self.stack.add_titled(manage_extensions_page, 'installed', _("Manage"))

        download_actions_page = DownloadSpicesPage(self, self.collection_type, self.spices, self.window)
        self.stack.add_titled(download_actions_page, 'more', _("Download"))

        if GLib.find_program_in_path("nemo-action-layout-editor"):
            for dir in GLib.get_system_data_dirs():
                path = Path(dir).joinpath("nemo/layout-editor")
                if path.exists():
                    sys.path.append(str(path))
                    try:
                        import nemo_action_layout_editor
                        editor = nemo_action_layout_editor.NemoActionsOrganizer(self.window)
                        editor.props.margin_start = 80
                        editor.props.margin_end = 80
                        editor.props.margin_top = 15
                        editor.props.margin_bottom = 30
                        self.stack.add_titled(editor, 'editor', _("Layout"))

                        def on_window_delete(window, event, data=None):
                            if not editor.quit():
                                return True

                        self.window.connect("delete-event", on_window_delete)
                    except Exception as e:
                        print(e)
                    break

class ManageActionsPage(ManageSpicesPage):
    directories = [f"{GLib.get_home_dir()}/.local/share/nemo/actions"]
    collection_type = "action"
    instance_button_text = _("Enable")
    remove_button_text = _("Disable")
    installed_page_title = _("Installed actions")
    uninstall_button_text = _("Uninstall")
    restore_button_text = _("Remove all")

    def __init__(self, parent, spices, window):
        super().__init__(parent, self.collection_type, spices, window)



#!/usr/bin/python3

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from xapp.SettingsWidgets import *
from bin.SettingsWidgets import SoundFileChooser, Keybinding

VARIABLE_TYPE_MAP = {
    "string"        :   str,
    "file"          :   str,
    "icon"          :   str,
    "sound"         :   str,
    "keybinding"    :   str,
    "integer"       :   int,
    "float"         :   float,
    "boolean"       :   bool
}

CLASS_TYPE_MAP = {
    "string"        :   Entry,
    "file"          :   FileChooser,
    "icon"          :   IconChooser,
    "sound"         :   SoundFileChooser,
    "keybinding"    :   Keybinding,
    "integer"       :   SpinButton,
    "float"         :   SpinButton,
    "boolean"       :   Switch
}

PROPERTIES_MAP = {
    "title"         : "label",
    "min"           : "mini",
    "max"           : "maxi",
    "step"          : "step",
    "units"         : "units",
    "select-dir"    : "dir_select",
    "expand-width"  : "expand_width"
}

def list_edit_factory(options):
    kwargs = {}
    if 'options' in options:
        kwargs['valtype'] = VARIABLE_TYPE_MAP[options['type']]
        widget_type = ComboBox
        options_list = options['options']
        if isinstance(options_list, dict):
            kwargs['options'] = [(b, a) for a, b in options_list.items()]
        else:
            kwargs['options'] = zip(options_list, options_list)
    else:
        widget_type = CLASS_TYPE_MAP[options["type"]]
    class Widget(widget_type):
        def __init__(self, **kwargs):
            super(Widget, self).__init__(**kwargs)

            if self.bind_dir is None:
                self.connect_widget_handlers()

        def get_range(self):
            return None

        def set_value(self, value):
            self.widget_value = value

        def get_value(self):
            if hasattr(self, "widget_value"):
                return self.widget_value
            else:
                return None

        def set_widget_value(self, value):
            if self.bind_dir is None:
                self.widget_value = value
                self.on_setting_changed()
            else:
                if hasattr(self, "bind_object"):
                    self.bind_object.set_property(self.bind_prop, value)
                else:
                    self.content_widget.set_property(self.bind_prop, value)

        def get_widget_value(self):
            if self.bind_dir is None:
                try:
                    return self.widget_value
                except Exception as e:
                    return None
            else:
                if hasattr(self, "bind_object"):
                    return self.bind_object.get_property(self.bind_prop)
                return self.content_widget.get_property(self.bind_prop)

    for prop in options:
        if prop in PROPERTIES_MAP:
            kwargs[PROPERTIES_MAP[prop]] = options[prop]

    return Widget(**kwargs)


class List(SettingsWidget):
    bind_dir = None

    def __init__(self, label=None, columns=None, height=200, size_group=None, \
                 dep_key=None, tooltip="", show_buttons=True, hidden_buttons=[]):
        super(List, self).__init__(dep_key=dep_key)
        self.columns = columns
        self.show_buttons = show_buttons
        self.hidden_buttons = hidden_buttons

        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_spacing(0)
        self.set_margin_left(0)
        self.set_margin_right(0)
        self.set_border_width(0)

        if label is not None:
            self.label = Gtk.Label(label)

        self.content_widget = Gtk.TreeView()

        scrollbox = Gtk.ScrolledWindow()
        scrollbox.set_size_request(-1, height)
        scrollbox.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.pack_start(scrollbox, True, True, 0)
        scrollbox.add(self.content_widget)

        types = []
        tv_columns = []
        for i in range(len(columns)):
            column_def = columns[i]
            types.append(VARIABLE_TYPE_MAP[column_def['type']])

            has_option_map = 'options' in column_def and isinstance(column_def['options'], dict)
            render_type = 'string' if has_option_map else column_def['type']

            if render_type == 'boolean':
                renderer = Gtk.CellRendererToggle()

                def toggle_checkbox(renderer, path, column):
                    self.model[path][column] = not self.model[path][column]
                    self.list_changed()

                renderer.connect('toggled', toggle_checkbox, i)
                prop_name = 'active'
            elif render_type == 'icon':
                renderer = Gtk.CellRendererPixbuf()
                prop_name = 'icon_name'
            else:
                renderer = Gtk.CellRendererText()
                prop_name = 'text'

            column = Gtk.TreeViewColumn(column_def['title'], renderer)

            if has_option_map:
                def map_func(col, rend, model, row_iter, data):
                    value = model[row_iter][data[1]]
                    for key, val in data[0].items():
                        if val == value:
                            rend.set_property('text', key)

                column.set_cell_data_func(renderer, map_func, [column_def['options'],i])
            else:
                column.add_attribute(renderer, prop_name, i)

            if 'align' in column_def:
                renderer.set_alignment(column_def['align'], 0.5)
                column.set_alignment(column_def['align'])

            column.set_resizable(True)
            self.content_widget.append_column(column)
        self.model = Gtk.ListStore(*types)
        self.content_widget.set_model(self.model)

        if show_buttons:
            button_toolbar = Gtk.Toolbar()
            button_toolbar.set_icon_size(1)
            Gtk.StyleContext.add_class(Gtk.Widget.get_style_context(button_toolbar), \
                                       "inline-toolbar")
            self.pack_start(button_toolbar, False, False, 0)

            self.add_button = Gtk.ToolButton(None, None)
            self.add_button.set_icon_name("xsi-list-add-symbolic")
            self.add_button.set_tooltip_text(_("Add new entry"))
            self.add_button.connect("clicked", self.add_item)
            self.remove_button = Gtk.ToolButton(None, None)
            self.remove_button.set_icon_name("xsi-list-remove-symbolic")
            self.remove_button.set_tooltip_text(_("Remove selected entry"))
            self.remove_button.connect("clicked", self.remove_item)
            self.remove_button.set_sensitive(False)
            self.edit_button = Gtk.ToolButton(None, None)
            self.edit_button.set_icon_name("xsi-list-edit-symbolic")
            self.edit_button.set_tooltip_text(_("Edit selected entry"))
            self.edit_button.connect("clicked", self.edit_item)
            self.edit_button.set_sensitive(False)
            self.move_up_button = Gtk.ToolButton(None, None)
            self.move_up_button.set_icon_name("xsi-go-up-symbolic")
            self.move_up_button.set_tooltip_text(_("Move selected entry up"))
            self.move_up_button.connect("clicked", self.move_item_up)
            self.move_up_button.set_sensitive(False)
            self.move_down_button = Gtk.ToolButton(None, None)
            self.move_down_button.set_icon_name("xsi-go-down-symbolic")
            self.move_down_button.set_tooltip_text(_("Move selected entry down"))
            self.move_down_button.connect("clicked", self.move_item_down)
            self.move_down_button.set_sensitive(False)
            if "+" not in self.hidden_buttons:
                button_toolbar.insert(self.add_button, 0)
            if "-" not in self.hidden_buttons:
                button_toolbar.insert(self.remove_button, 1)
            if "edit" not in self.hidden_buttons:
                button_toolbar.insert(self.edit_button, 2)
            if "up" not in self.hidden_buttons:
                button_toolbar.insert(self.move_up_button, 3)
            if "down" not in self.hidden_buttons:
                button_toolbar.insert(self.move_down_button, 4)

        self.content_widget.get_selection().connect("changed", self.update_button_sensitivity)
        self.content_widget.set_activate_on_single_click(False)
        self.content_widget.connect("row-activated", self.on_row_activated)

        self.set_tooltip_text(tooltip)

    def update_button_sensitivity(self, *args):
        if not self.show_buttons:
            return
        model, selected = self.content_widget.get_selection().get_selected()
        if selected is None:
            self.remove_button.set_sensitive(False)
            self.edit_button.set_sensitive(False)
        else:
            self.remove_button.set_sensitive(True)
            self.edit_button.set_sensitive(True)

        if selected is None or model.iter_previous(selected) is None:
            self.move_up_button.set_sensitive(False)
        else:
            self.move_up_button.set_sensitive(True)

        if selected is None or model.iter_next(selected) is None:
            self.move_down_button.set_sensitive(False)
        else:
            self.move_down_button.set_sensitive(True)

    def on_row_activated(self, *args):
        if "edit" not in self.hidden_buttons:
            self.edit_item()

    def add_item(self, *args):
        data = self.open_add_edit_dialog()
        if data is not None:
            self.model.append(data)
            self.list_changed()

    def remove_item(self, *args):
        model, t_iter = self.content_widget.get_selection().get_selected()
        model.remove(t_iter)

        self.list_changed()

    def edit_item(self, *args):
        model, t_iter = self.content_widget.get_selection().get_selected()
        data = self.open_add_edit_dialog(model[t_iter])
        if data is not None:
            for i in range(len(data)):
                self.model[t_iter][i] = data[i]
            self.list_changed()

    def move_item_up(self, *args):
        model, t_iter = self.content_widget.get_selection().get_selected()
        model.swap(t_iter, model.iter_previous(t_iter))
        self.list_changed()

    def move_item_down(self, *args):
        model, t_iter = self.content_widget.get_selection().get_selected()
        model.swap(t_iter, model.iter_next(t_iter))
        self.list_changed()

    def open_add_edit_dialog(self, info=None):
        if info is None:
            title = _("Add new entry")
        else:
            title = _("Edit entry")
        dialog = Gtk.Dialog(title, self.get_toplevel(), Gtk.DialogFlags.MODAL,
                            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                             Gtk.STOCK_OK, Gtk.ResponseType.OK))

        content_area = dialog.get_content_area()
        content_area.set_margin_right(30)
        content_area.set_margin_left(30)
        content_area.set_margin_top(20)
        content_area.set_margin_bottom(20)

        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.IN)
        frame_style = frame.get_style_context()
        frame_style.add_class("view")
        content_area.add(frame)

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        frame.add(content)

        widgets = []
        for i in range(len(self.columns)):
            if len(widgets) != 0:
                content.add(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))

            widget = list_edit_factory(self.columns[i])
            widgets.append(widget)

            settings_box = Gtk.ListBox()
            settings_box.set_selection_mode(Gtk.SelectionMode.NONE)

            content.pack_start(settings_box, True, True, 0)
            settings_box.add(widget)

            if info is not None and info[i] is not None:
                widget.set_widget_value(info[i])
            elif "default" in self.columns[i]:
                widget.set_widget_value(self.columns[i]["default"])

        content_area.show_all()
        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            values = []
            for widget in widgets:
                values.append(widget.get_widget_value())

            dialog.destroy()
            return values

        dialog.destroy()
        return None

    def list_changed(self, *args):
        data = []
        for row in self.model:
            i = 0
            row_info = {}
            for column in self.columns:
                row_info[column["id"]] = row[i]
                i += 1
            data.append(row_info)

        self.set_value(data)
        self.update_button_sensitivity()

    def on_setting_changed(self, *args):
        self.model.clear()
        rows = self.get_value()
        for row in rows:
            row_info = []
            for column in self.columns:
                cid = column["id"]
                if cid in row:
                    row_info.append(row[column["id"]])
                elif "default" in column:
                    row_info.append(column["default"])
                else:
                    row_info.append(None)
            self.model.append(row_info)

        self.content_widget.columns_autosize()

    def connect_widget_handlers(self, *args):
        pass



import os
import sys

import gi
gi.require_version('XApp', '1.0')
from gi.repository import GdkPixbuf, Gio, GLib, GObject, XApp

from backends.base import Backend


BUS_NAME = 'org.cinnamon.Screenshot'
OBJECT_PATH = '/org/cinnamon/Screenshot'
INTERFACE = 'org.cinnamon.Screenshot'


class CinnamonBackend(Backend, GObject.Object):
    __gsignals__ = {
        'online-changed': (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self):
        GObject.Object.__init__(self)
        self._proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
            None,
            BUS_NAME, OBJECT_PATH, INTERFACE,
            None,
        )
        self._proxy.connect('notify::g-name-owner', self._on_name_owner_changed)

    def _on_name_owner_changed(self, *_args):
        self.emit('online-changed', self.is_available())

    def is_available(self):
        return self._proxy.get_name_owner() is not None

    def _tempfile(self):
        return os.path.join(XApp.get_tmp_dir(), f'cinnamon-screenshot-{os.getpid()}.png')

    def _call(self, method, params, on_result):
        """Invoke a DBus method asynchronously; deliver the unpacked
        result tuple (or None on failure) to on_result."""
        def cb(proxy, res):
            try:
                result = proxy.call_finish(res).unpack()
            except GLib.Error as exc:
                print(f'cinnamon-screenshot: DBus {method} failed: {exc.message}', file=sys.stderr)
                result = None
            on_result(result)
        self._proxy.call(method, params, Gio.DBusCallFlags.NONE, -1, None, cb)

    def _load_and_unlink(self, path):
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
        except GLib.Error:
            pixbuf = None
        try:
            os.unlink(path)
        except OSError:
            pass
        return pixbuf

    def _deliver_image(self, path, result, on_done):
        if not result or not result[0]:
            on_done(None)
            return
        on_done(self._load_and_unlink(result[1] or path))

    def screenshot(self, include_pointer, on_done):
        path = self._tempfile()
        self._call('Screenshot',
                   GLib.Variant('(bs)', (include_pointer, path)),
                   lambda result: self._deliver_image(path, result, on_done))

    def screenshot_window(self, include_pointer, include_shadow, on_done):
        path = self._tempfile()
        self._call('ScreenshotWindow',
                   GLib.Variant('(bbs)', (include_shadow, include_pointer, path)),
                   lambda result: self._deliver_image(path, result, on_done))

    def screenshot_window_by_id(self, window_id, include_pointer, include_shadow, on_done):
        path = self._tempfile()
        self._call('ScreenshotWindowById',
                   GLib.Variant('(tbbs)', (window_id, include_shadow, include_pointer, path)),
                   lambda result: self._deliver_image(path, result, on_done))

    def screenshot_area(self, x, y, w, h, include_pointer, on_done):
        path = self._tempfile()
        self._call('ScreenshotArea',
                   GLib.Variant('(iiiibs)', (x, y, w, h, include_pointer, path)),
                   lambda result: self._deliver_image(path, result, on_done))

    def flash_area(self, x, y, w, h):
        self._proxy.call('FlashArea', GLib.Variant('(iiii)', (x, y, w, h)),
                         Gio.DBusCallFlags.NONE, -1, None, None)

    def select_area(self, on_done):
        self._call('SelectArea', None,
                   lambda result: on_done(tuple(result) if result else None))

    def select_window(self, on_done):
        self._call('SelectWindow', None,
                   lambda result: on_done(result[0] if result else None))



import os
import sys

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, Gio, GLib, Gtk


_FILE_MANAGER1_BUS = 'org.freedesktop.FileManager1'
_FILE_MANAGER1_PATH = '/org/freedesktop/FileManager1'

_PIXBUF_FORMAT_FOR_EXT = {
    '.png':  ('png',  []),
    '.jpg':  ('jpeg', [('quality', '95')]),
    '.jpeg': ('jpeg', [('quality', '95')]),
    '.bmp':  ('bmp',  []),
    '.tif':  ('tiff', []),
    '.tiff': ('tiff', []),
}

def format_for_path(path):
    ext = os.path.splitext(path)[1].lower()
    return _PIXBUF_FORMAT_FOR_EXT.get(ext, _PIXBUF_FORMAT_FOR_EXT['.png'])

def save_pixbuf(pixbuf, path):
    fmt, options = format_for_path(path)
    keys = [k for k, _ in options]
    values = [v for _, v in options]
    pixbuf.savev(path, fmt, keys, values)

def copy_pixbuf_to_clipboard(pixbuf):
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_image(pixbuf)
    clipboard.store()

def monitor_rect(idx):
    """Return (x, y, w, h) of the monitor at the given index, or None if
    the index is out of range or there's only one monitor (caller should
    treat that as 'use the whole screen')."""
    if idx is None:
        return None
    display = Gdk.Display.get_default()
    if display is None:
        return None
    n = display.get_n_monitors()
    if n <= 1 or idx < 0 or idx >= n:
        return None
    geom = display.get_monitor(idx).get_geometry()
    return (geom.x, geom.y, geom.width, geom.height)

def monitor_rect_for_window(gdk_window):
    """Return (x, y, w, h) of the monitor the given GdkWindow is on,
    or None if it can't be determined."""
    if gdk_window is None:
        return None
    display = Gdk.Display.get_default()
    if display is None:
        return None
    monitor = display.get_monitor_at_window(gdk_window)
    if monitor is None:
        return None
    geom = monitor.get_geometry()
    return (geom.x, geom.y, geom.width, geom.height)

def build_filename(directory, file_type='png'):
    """Build a 'Screenshot <iso-timestamp>.<ext>' path inside directory.
    Caller is responsible for passing an existing directory."""
    timestamp = GLib.DateTime.new_now_local().format('%Y-%m-%d %H-%M-%S.%f')[:-3]
    return os.path.join(directory, f'{_('Screenshot')} {timestamp}.{file_type}')

def show_in_file_manager(uri):
    f = Gio.File.new_for_uri(uri)
    is_dir = False
    try:
        info = f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_TYPE,
                            Gio.FileQueryInfoFlags.NONE, None)
        is_dir = info.get_file_type() == Gio.FileType.DIRECTORY
    except GLib.Error:
        pass

    if not is_dir:
        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            bus.call_sync(
                _FILE_MANAGER1_BUS, _FILE_MANAGER1_PATH, _FILE_MANAGER1_BUS,
                'ShowItems',
                GLib.Variant('(ass)', ([uri], '')),
                None, Gio.DBusCallFlags.NONE, -1, None,
            )
            return
        except GLib.Error:
            pass

    target = uri if is_dir else (f.get_parent() or f).get_uri()
    try:
        Gio.AppInfo.launch_default_for_uri(target, None)
    except GLib.Error as exc:
        print(f'cinnamon-screenshot: failed to open file manager: {exc.message}',
              file=sys.stderr)



#!/usr/bin/python3

import os
import gettext
from mintcommon import additionalfiles

DOMAIN = "cinnamon"
PATH = "/usr/share/locale"

os.environ['LANGUAGE'] = "en_US.UTF-8"
gettext.install(DOMAIN, PATH)

prefix = """[Desktop Entry]
Exec=env WEBKIT_DISABLE_COMPOSITING_MODE=1 cinnamon-settings
Icon=preferences-desktop
Terminal=false
Type=Application
Categories=Settings;
StartupNotify=false
OnlyShowIn=X-Cinnamon;
"""

additionalfiles.generate(DOMAIN, PATH, "files/usr/share/applications/cinnamon-settings.desktop", prefix,
    _("System Settings"),
    _("Control Center"),
    "",
    keywords=_("Preferences,Settings"))

prefix = """[Desktop Entry]
Exec=cinnamon-settings-users
Icon=system-users
Terminal=false
Type=Application
Categories=System;Settings;
StartupNotify=false
OnlyShowIn=X-Cinnamon;
"""

additionalfiles.generate(DOMAIN, PATH, "files/usr/share/applications/cinnamon-settings-users.desktop", prefix,
    _("Users and Groups"),
    _("Add or remove users and groups"),
    "",
    keywords=_("Preferences,Settings,Users,Groups"))

prefix = """[Desktop Entry]
Exec=cinnamon-dbus-command ToggleKeyboard
Icon=cinnamon-virtual-keyboard
Terminal=false
Type=Application
Categories=Utility;
OnlyShowIn=X-Cinnamon;
"""

additionalfiles.generate(DOMAIN, PATH, "files/usr/share/applications/cinnamon-onscreen-keyboard.desktop", prefix,
    _("Virtual keyboard"),
    _("Turn on-screen keyboard on or off"),
    "",
    keywords=_("onboard,keyboard,caribou"))

SCREENSHOT_DESKTOP = "files/usr/share/applications/cinnamon-screenshot.desktop"

prefix = """[Desktop Entry]
Exec=cinnamon-screenshot --interactive
Terminal=false
Type=Application
Icon=applets-screenshooter
StartupNotify=true
Categories=GTK;Utility;
OnlyShowIn=X-Cinnamon;
Actions=screen-shot;window-shot;area-shot;
"""

additionalfiles.generate(DOMAIN, PATH, SCREENSHOT_DESKTOP, prefix,
    _("Screenshot"),
    _("Save images of your screen or individual windows"),
    "",
    keywords=_("snapshot,capture,print,screenshot"))

additionalfiles.generate(DOMAIN, PATH, SCREENSHOT_DESKTOP, "\n[Desktop Action screen-shot]\n",
    _("Take a Screenshot of the Whole Screen"),
    None,
    "Exec=cinnamon-screenshot\n",
    append=True)

additionalfiles.generate(DOMAIN, PATH, SCREENSHOT_DESKTOP, "\n[Desktop Action window-shot]\n",
    _("Take a Screenshot of the Current Window"),
    None,
    "Exec=cinnamon-screenshot -w\n",
    append=True)

additionalfiles.generate(DOMAIN, PATH, SCREENSHOT_DESKTOP, "\n[Desktop Action area-shot]\n",
    _("Take a Screenshot of an Area"),
    None,
    "Exec=cinnamon-screenshot -a\n",
    append=True)



#!/usr/bin/python3

from gi.repository import Gtk, GObject, GLib, Gdk, GdkPixbuf, Gio
import os
import gettext
import datetime
gettext.install("cinnamon", "/usr/share/locale")

class BaseChooserButton(Gtk.Button):
    def __init__ (self, has_button_label=False, frame=False):
        super(BaseChooserButton, self).__init__()
        self.has_button_label = has_button_label
        self.frame = frame
        self.set_valign(Gtk.Align.CENTER)
        self.menu = Gtk.Menu()
        self.button_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=2)
        self.button_image = Gtk.Image()

        if self.frame:
            f = Gtk.Frame(valign=Gtk.Align.END, halign=Gtk.Align.CENTER)
            f.add(self.button_image)
            self.button_box.add(f)
        else:
            self.button_box.add(self.button_image)

        if self.has_button_label:
            self.button_label = Gtk.Label()
            self.button_box.add(self.button_label)
        self.add(self.button_box)
        self.connect("button-release-event", self._on_button_clicked)

    def popup_menu_below_button (self, *args):
        # the introspection for GtkMenuPositionFunc seems to change with each Gtk version,
        # this is a workaround to make sure we get the menu and the widget
        menu = args[0]
        widget = args[-1]
        window = widget.get_window()
        screen = window.get_screen()
        monitor = screen.get_monitor_at_window(window)

        warea = screen.get_monitor_workarea(monitor)
        wrect = widget.get_allocation()
        mrect = menu.get_allocation()

        unused_var, window_x, window_y = window.get_origin()

        # Position left edge of the menu with the right edge of the button
        x = window_x + wrect.x + wrect.width
        # Center the menu vertically with respect to the monitor
        y = warea.y + (warea.height / 2) - (mrect.height / 2)

        # Now, check if we're still touching the button - we want the right edge
        # of the button always 100% touching the menu

        if y > (window_y + wrect.y):
            y = y - (y - (window_y + wrect.y))
        elif (y + mrect.height) < (window_y + wrect.y + wrect.height):
            y = y + ((window_y + wrect.y + wrect.height) - (y + mrect.height))

        push_in = True # push_in is True so all menu is always inside screen
        return x, y, push_in

    def _on_button_clicked(self, widget, event):
        if event.button == 1:
            self.menu.show_all()
            self.menu.popup(None, None, self.popup_menu_below_button, self, event.button, event.time)

class PictureChooserButton(BaseChooserButton):
    def __init__ (self, num_cols=4, button_picture_width=24, menu_picture_width=24, has_button_label=False, keep_square=False, frame=False):
        super(PictureChooserButton, self).__init__(has_button_label, frame)
        self.num_cols = num_cols
        self.scale = self.get_scale_factor()
        self.button_picture_width = button_picture_width
        self.menu_picture_width = menu_picture_width
        self.keep_square = keep_square
        self.row = 0
        self.col = 0
        self.progress = 0.0

        context = self.get_style_context()
        context.add_class("gtkstyle-fallback")

        self.button_image.set_valign(Gtk.Align.CENTER)
        if self.keep_square:
            self.button_image.set_size_request(button_picture_width / self.scale, button_picture_width / self.scale)
        else:
            self.button_image.set_size_request(button_picture_width / self.scale, -1)

        self.connect_after("draw", self.on_draw)

    def on_draw(self, widget, cr, data=None):
        if self.progress == 0:
            return False
        box = widget.get_allocation()

        context = widget.get_style_context()
        c = context.get_background_color(Gtk.StateFlags.SELECTED)

        max_length = box.width * .6
        start = (box.width - max_length) / 2
        y = box.height - 5

        cr.save()

        cr.set_source_rgba(c.red, c.green, c.blue, c.alpha)
        cr.set_line_width(3)
        cr.set_line_cap(1)
        cr.move_to(start, y)
        cr.line_to(start + (self.progress * max_length), y)
        cr.stroke()

        cr.restore()
        return False

    def increment_loading_progress(self, inc):
        progress = self.progress + inc
        self.progress = min(1.0, progress)
        self.queue_draw()

    def reset_loading_progress(self):
        self.progress = 0.0
        self.queue_draw()

    def create_scaled_surface(self, path):
        w = self.button_picture_width * self.scale
        h = -1 if not self.keep_square else self.button_picture_width * self.scale

        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, w, h)
            if pixbuf:
                return Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.scale)
        except GLib.Error as e:
            print(f"Could not load thumbnail file '{path}': {e.message}")
        return None

    def set_picture_from_file (self, path):
        surface = self.create_scaled_surface(path)
        if surface:
            self.button_image.set_from_surface(surface)
        else:
            self.button_image.set_from_icon_name("user-generic", Gtk.IconSize.BUTTON)

    def set_button_label(self, label):
        if self.has_button_label:
            self.button_label.set_markup(label)

    def _on_picture_selected(self, menuitem, path, callback, id=None):
        if id is not None:
            result = callback(path, id)
        else:
            result = callback(path)

        if result:
            self.set_picture_from_file(path)

    def clear_menu(self):
        menu = self.menu
        self.menu = Gtk.Menu()
        self.row = 0
        self.col = 0
        menu.destroy()

    def add_picture(self, path, callback, title=None, id=None):
        image = Gtk.Image()
        image.set_size_request(self.menu_picture_width / self.scale, -1)

        surface = self.create_scaled_surface(path)

        if surface:
            image.set_from_surface(surface)
        else:
            image.set_from_icon_name("user-generic", Gtk.IconSize.BUTTON)

        if self.frame:
            frame = Gtk.Frame(halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER)
            frame.add(image)
            menu_image = frame
        else:
            menu_image = image

        menuitem = Gtk.MenuItem()
        if title is not None:
            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=2, valign=Gtk.Align.END)
            vbox.add(menu_image)
            label = Gtk.Label()
            label.set_text(title)
            vbox.add(label)
            menuitem.add(vbox)
        else:
            menuitem.add(menu_image)
        if id is not None:
            menuitem.connect('activate', self._on_picture_selected, path, callback, id)
        else:
            menuitem.connect('activate', self._on_picture_selected, path, callback)
        self.menu.attach(menuitem, self.col, self.col+1, self.row, self.row+1)
        self.col = (self.col+1) % self.num_cols
        if self.col == 0:
            self.row = self.row + 1

    def add_separator(self, label_text=None):
        if self.col > 0:
            self.row = self.row + 1
            self.col = 0

        if label_text is not None:
            item = Gtk.MenuItem()
            item.set_sensitive(False)
            # Add centered, bold label
            label = Gtk.Label()
            label.set_markup(f"<b>{label_text}</b>")
            label.set_halign(Gtk.Align.START)
            item.add(label)
        else:
            item = Gtk.SeparatorMenuItem()

        self.menu.attach(item, 0, self.num_cols, self.row, self.row + 1)
        self.row = self.row + 1

    def add_menuitem(self, menuitem):
        if self.col > 0:
            self.row = self.row + 1
            self.col = 0

        self.menu.attach(menuitem, 0, self.num_cols, self.row, self.row+1)
        self.row = self.row + 1

class DateChooserButton(Gtk.Button):
    __gsignals__ = {
        'date-changed': (GObject.SignalFlags.RUN_FIRST, None, (int,int,int))
    }

    def __init__(self, follow_current=False, date=None):
        super(DateChooserButton, self).__init__()

        if not follow_current and date is not None:
            self.set_date(date)
        else:
            self.set_date(datetime.date.today())

        if follow_current:
            GLib.timeout_add_seconds(1, self.update_date)

        self.connect("clicked", self.on_button_clicked)

    def on_button_clicked(self, *args):
        self.dialog = Gtk.Dialog(transient_for=self.get_toplevel(),
                                 title=_("Select a date"),
                                 flags=Gtk.DialogFlags.MODAL,
                                 buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                          Gtk.STOCK_OK, Gtk.ResponseType.OK))

        content = self.dialog.get_content_area()
        content.props.margin_start = 20
        content.props.margin_end = 20
        content.props.margin_top = 20
        content.props.margin_bottom = 20
        content.props.spacing = 15

        today = Gtk.Button(label=_("Today"))
        today.props.halign = Gtk.Align.CENTER
        content.pack_start(today, False, False, 0)

        calendar = Gtk.Calendar()
        content.pack_start(calendar, True, True, 0)
        calendar.select_month(self.date.month-1, self.date.year)
        calendar.select_day(self.date.day)

        def select_today(*args):
            date = datetime.date.today()
            calendar.select_month(date.month - 1, date.year)
            calendar.select_day(date.day)
        today.connect("clicked", select_today)

        content.show_all()

        response = self.dialog.run()

        if response == Gtk.ResponseType.OK:
            date = calendar.get_date()
            self.set_date(datetime.date(date[0], date[1]+1, date[2])) # Gtk.Calendar uses 0 based month
            self.emit("date-changed", self.date.year, self.date.month, self.date.day)

        self.dialog.destroy()

    def update_date(self, *args):
        self.set_date(datetime.date.today())
        return True

    def get_date(self):
        return self.date

    def set_date(self, date):
        """Sets the date of the widget.
        Date can be a date or datetime class from the datetime module or a (y,m,d) tuple.
        """
        if isinstance(date, datetime.date):
            self.date = date
        elif isinstance(date, datetime.datetime):
            self.date = date.date()
        elif isinstance(date, tuple):
            self.date = datetime.date(*date)
        else:
            raise ValueError('Invalid date format. Date must be of type datetime.date, datetime.datetime, or a tuple of the form (year, month, day)')

        # Translators: This is the date which appears in the calendar
        # applet, as a tooltip and as a title header inside the applet.
        # The format uses the strftime syntax.
        # %A, %B %-e, %Y is the American Date format (Saturday, January 4, 2020):
        # %A is the name of the day, %B is the name of the month
        # %-e is the non-padded day number, %Y is the year.
        # for info on the strftime format visit: http://man7.org/linux/man-pages/man3/strftime.3.html
        # for info on international date formats visit: http://www.localeplanet.com/compare/date-pattern.html?pat=FULL
        date_string = self.date.strftime(_("%A, %B %-e, %Y"))
        self.set_label(date_string)

class TimeChooserButton(Gtk.Button):
    __gsignals__ = {
        'time-changed': (GObject.SignalFlags.RUN_FIRST, None, (int,int,int))
    }

    def __init__(self, follow_current=False, time=None, show_seconds='default'):
        super(TimeChooserButton, self).__init__()

        if show_seconds == 'default':
            self.show_seconds_override_default = False
        else:
            self.show_seconds_override_default = True
            if show_seconds == 'true':
                self.show_seconds = True
            elif show_seconds == 'false':
                self.show_seconds = False
            else:
                raise ValueError('Invalid argument: show_seconds must be default, true, or false')

        self.settings = Gio.Settings.new('org.cinnamon.desktop.interface')

        if not follow_current and time is not None:
            self.set_time(time)
        else:
            self.set_time(datetime.datetime.now().time())

        if follow_current:
            GLib.timeout_add_seconds(1, self.update_time)

        self.connect('clicked', self.on_button_clicked)

    def get_uses_seconds(self):
        if self.show_seconds_override_default:
            return self.show_seconds
        else:
            return self.settings.get_boolean('clock-show-seconds')

    def on_button_clicked(self, *args):
        dialog = TimeChooserDialog(self.time, self.get_toplevel(), self.get_uses_seconds(), self.settings.get_boolean('clock-use-24h'))

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self.set_time(dialog.get_time())
            self.emit("time-changed", self.time.hour, self.time.minute, self.time.second)

        dialog.destroy()

    def update_time(self, *args):
        self.set_time(datetime.datetime.now().time())
        return True

    def get_time(self):
        return self.time

    def set_time(self, time):
        """Sets the time of the widget.
        Time can be a time or datetime class from the datetime module or a (h,m) or (h,m,s) tuple.
        """
        if isinstance(time, datetime.time):
            self.time = time
        elif isinstance(time, datetime.datetime):
            self.time = time.time()
        elif isinstance(time, tuple):
            self.time = datetime.time(*time)
        else:
            raise ValueError('Invalid time format. Must be of type datetime.time, datetime.datetime, or a tuple of the form (hour, minute[, second])')

        self.update_label()

    def update_label(self):
        if self.settings.get_boolean('clock-use-24h'):
            if self.get_uses_seconds():
                format_code = gettext.dgettext('cinnamon-desktop', '%R:%S')
            else:
                format_code = gettext.dgettext('cinnamon-desktop', '%R')
        else:
            if self.get_uses_seconds():
                format_code = gettext.dgettext('cinnamon-desktop', '%l:%M:%S %p')
            else:
                format_code = gettext.dgettext('cinnamon-desktop', '%l:%M %p')
        time_string = self.time.strftime(format_code)
        self.set_label(time_string)

class TimeChooserDialog(Gtk.Dialog):
    def __init__(self, time, window, use_seconds, use24hour):
        super(TimeChooserDialog, self).__init__(title = _("Select a time"),
                                                transient_for = window,
                                                flags = Gtk.DialogFlags.MODAL,
                                                buttons = (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                                         Gtk.STOCK_OK, Gtk.ResponseType.OK))

        self.time = {'hour': time.hour, 'minute': time.minute, 'second': time.second}
        self.use_seconds = use_seconds
        self.use24hour = use24hour
        self.markup = lambda text: f'<span weight="bold" size="xx-large">{text}</span>'

        content = self.get_content_area()

        grid = Gtk.Grid(halign=Gtk.Align.CENTER)
        content.pack_start(grid, False, False, 0)

        grid.attach(Gtk.Label(self.markup(_("Hour")), use_markup=True), 0, 0, 1, 1)
        grid.attach(Gtk.Label(self.markup(':'), use_markup=True), 1, 2, 1, 1)
        grid.attach(Gtk.Label(self.markup(_("Minute")), use_markup=True), 2, 0, 1, 1)

        unit_defs = [('hour', 0), ('minute', 2)]

        if self.use_seconds:
            grid.attach(Gtk.Label(self.markup(':'), use_markup=True), 3, 2, 1, 1)
            grid.attach(Gtk.Label(self.markup(_("Second")), use_markup=True), 4, 0, 1, 1)
            unit_defs.append(('second', 4))

        self.labels = {}
        for ttype, column in unit_defs:
            self.labels[ttype] = Gtk.Label(self.markup(self.time[ttype]), use_markup=True)
            grid.attach(self.labels[ttype], column, 2, 1, 1)

            up_button = Gtk.Button.new_from_icon_name('xsi-pan-up-symbolic', Gtk.IconSize.DIALOG)
            down_button = Gtk.Button.new_from_icon_name('xsi-pan-down-symbolic', Gtk.IconSize.DIALOG)
            up_button.set_relief(Gtk.ReliefStyle.NONE)
            down_button.set_relief(Gtk.ReliefStyle.NONE)
            grid.attach(up_button, column, 1, 1, 1)
            grid.attach(down_button, column, 3, 1, 1)
            up_button.connect('clicked', self.shift_time, ttype, 1)
            down_button.connect('clicked', self.shift_time, ttype, -1)

        content.show_all()

    def shift_time(self, button, ttype, offset):
        self.time[ttype] += offset
        self.time['hour'] = self.time['hour'] % 24
        self.time['minute'] = self.time['minute'] % 60
        self.time['second'] = self.time['second'] % 60
        self.labels[ttype].set_label(self.markup(self.time[ttype]))

    def get_time(self):
        return datetime.time(self.time['hour'], self.time['minute'], self.time['second'])



#!/usr/bin/python3

from gi.repository import Gio, Gtk, CinnamonDesktop, Pango, GLib

from xapp.SettingsWidgets import SettingsSection

class XkbSettingsEditor(SettingsSection):
    # Groups that are hidden from UI (broken or not supported in Cinnamon)
    FORBIDDEN_GROUPS = ['custom']

    # Allowed options per group. If a group is in this dict, only the listed options are shown.
    # If a group is not in this dict, all options from XkbInfo are shown.
    ALLOWED_OPTIONS = {
        'grp': {
            'wayland_allowed': False,
            'options':
                [
                    "toggle",
                    "lalt_toggle",
                    "lwin_toggle",
                    "rwin_toggle",
                    "lshift_toggle",
                    "rshift_toggle",
                    "lctrl_toggle",
                    "rctrl_toggle",
                    "sclk_toggle",
                    "menu_toggle",
                    "caps_toggle",
                    "shift_caps_toggle",
                    "shifts_toggle",
                    "alt_caps_toggle",
                    "alt_space_toggle",
                    "win_space_toggle",
                    "ctrl_space_toggle",
                    "ctrl_shift_toggle",
                    "lctrl_lshift_toggle",
                    "rctrl_rshift_toggle",
                    "ctrl_alt_toggle",
                    "alt_shift_toggle",
                    "lalt_lshift_toggle",
                ]
        }
    }

    LEFT_PANE_WIDTH = 250
    LEFT_PANE_MIN_WIDTH = 150

    def __init__(self, **kwargs):
        super().__init__()
        self.settings = Gio.Settings(schema_id="org.cinnamon.desktop.input-sources")
        self.xkb_info = CinnamonDesktop.XkbInfo.new_with_extras()

        self._building = False
        self._internal_write = False
        self._current_group = None
        self._rows_by_group_id = {}

        self.groups = self._build_groups()

        self._build_ui()
        self._populate_group_list()

        self.settings.connect("changed::xkb-options", self._on_external_change)

        first = self.groups_listbox.get_row_at_index(0)
        if first is not None:
            self.groups_listbox.select_row(first)

    def _is_group_allowed(self, group_id):
        return group_id not in self.FORBIDDEN_GROUPS

    def _build_groups(self):
        """Build an ordered list of group descriptors:
        { id, description, multi, all_ids, displayed:[(option_id, description)] }
        'id' is the XkbInfo group id and is used for every XkbInfo lookup.
        'all_ids' is the full set of the group's option ids (including ones not
        offered in the UI) and is used to map a stored option back to its group.
        Mapping is by exact option-id membership, not by prefix: a prefix can
        span two groups (e.g. 'keypad:legacy' is in the keypad group while
        'keypad:pointerkeys' is in compat), but the option ids never overlap.
        """
        groups = []
        for group_id in self.xkb_info.get_all_option_groups():
            if not self._is_group_allowed(group_id):
                continue

            description = self.xkb_info.description_for_group(group_id)
            if not description:
                continue

            # Tag groups whose options only work under X11 (e.g. muffin has no
            # 'grp' switching support on Wayland), driven by 'wayland_allowed'.
            group_config = self.ALLOWED_OPTIONS.get(group_id)
            if group_config is not None and not group_config.get('wayland_allowed', True):
                description = _("%s (X11 only)") % description

            all_options = self.xkb_info.get_options_for_group(group_id) or []
            if not all_options:
                continue

            # Options actually offered in the UI (curated where a whitelist exists).
            if group_id in self.ALLOWED_OPTIONS:
                allowed = set(self.ALLOWED_OPTIONS[group_id]['options'])
                displayed_ids = [o for o in all_options
                                 if ':' in o and o.split(':', 1)[1] in allowed]
            else:
                displayed_ids = list(all_options)

            multi = self.xkb_info.get_option_group_allows_multiple_selection(group_id)
            # Muffin only supports a single layout-switch toggle.
            if group_id == "grp":
                multi = False

            groups.append({
                "id": group_id,
                "description": description,
                "multi": multi,
                "all_ids": set(all_options),
                "displayed": sorted(
                    [(o, self.xkb_info.description_for_option(group_id, o) or o)
                     for o in displayed_ids],
                    key=lambda item: item[1].lower()),
            })

        groups.sort(key=lambda g: g["description"].lower())
        return groups

    def _stored_options(self):
        return self.settings.get_strv("xkb-options")

    def _stored_for_group(self, group):
        return [o for o in self._stored_options() if o in group["all_ids"]]

    def _option_description(self, group, option_id):
        return self.xkb_info.description_for_option(group["id"], option_id) or option_id

    def _sort_group_rows(self, row1, row2):
        # Groups with an active option sort ahead of unset ones, alphabetical by
        # description within each tier.
        active1 = bool(self._stored_for_group(row1.group))
        active2 = bool(self._stored_for_group(row2.group))
        if active1 != active2:
            return -1 if active1 else 1

        d1 = row1.group["description"].lower()
        d2 = row2.group["description"].lower()
        return (d1 > d2) - (d1 < d2)

    def _update_header(self, row, before):
        # A separator marks the boundary between the active groups and the
        # unset ones.
        if before is not None and \
           self._stored_for_group(before.group) and not self._stored_for_group(row.group):
            separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            separator.set_margin_top(4)
            separator.set_margin_bottom(4)
            row.set_header(separator)
        else:
            row.set_header(None)

    def _resort(self):
        self.groups_listbox.invalidate_sort()
        self.groups_listbox.invalidate_headers()

    # ------------------------------------------------------------------- view

    def _build_ui(self):
        paned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL, wide_handle=True)
        paned.set_position(self.LEFT_PANE_WIDTH)

        self.groups_listbox = Gtk.ListBox()
        self.groups_listbox.set_selection_mode(Gtk.SelectionMode.BROWSE)
        self.groups_listbox.set_sort_func(self._sort_group_rows)
        self.groups_listbox.set_header_func(self._update_header)
        self.groups_listbox.connect("row-selected", self._on_group_selected)

        left_scrolled = Gtk.ScrolledWindow()
        left_scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        left_scrolled.set_shadow_type(Gtk.ShadowType.IN)
        left_scrolled.set_size_request(self.LEFT_PANE_MIN_WIDTH, -1)
        left_scrolled.add(self.groups_listbox)

        self.detail_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=2)
        self.detail_box.props.margin = 4

        right_scrolled = Gtk.ScrolledWindow()
        right_scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        right_scrolled.set_shadow_type(Gtk.ShadowType.IN)
        right_scrolled.set_min_content_height(300)
        right_scrolled.set_min_content_width(200)
        right_scrolled.add(self.detail_box)

        paned.pack1(left_scrolled, False, False)
        paned.pack2(right_scrolled, True, True)

        self.pack_start(paned, True, True, 0)

    def _populate_group_list(self):
        bold = Pango.AttrList()
        bold.insert(Pango.attr_weight_new(Pango.Weight.BOLD))

        for group in self.groups:
            title = Gtk.Label(label=group["description"], xalign=0)
            title.set_attributes(bold)
            title.set_ellipsize(Pango.EllipsizeMode.END)

            subtitle = Gtk.Label(xalign=0)
            subtitle.get_style_context().add_class("dim-label")
            subtitle.set_ellipsize(Pango.EllipsizeMode.END)
            subtitle.set_no_show_all(True)

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=2, margin_left=2)
            vbox.pack_start(title, False, False, 0)
            vbox.pack_start(subtitle, False, False, 0)

            row = Gtk.ListBoxRow()
            row.group = group
            row.subtitle = subtitle
            row.add(vbox)

            self.groups_listbox.add(row)
            self._rows_by_group_id[group["id"]] = row
            self._update_subtitle(group)

        self.groups_listbox.show_all()

    def _build_detail(self, group):
        self._building = True

        for child in self.detail_box.get_children():
            self.detail_box.remove(child)

        active = self._stored_for_group(group)

        if group["multi"]:
            self._build_multi(group, active)
        else:
            self._build_single(group, active)

        self.detail_box.show_all()
        self._building = False

    def _build_multi(self, group, active):
        shown = set()
        for option_id, description in group["displayed"]:
            check = Gtk.CheckButton(label=description)
            check.set_active(option_id in active)
            check.connect("toggled", self._on_check_toggled, group, option_id)
            self.detail_box.pack_start(check, False, False, 0)
            shown.add(option_id)

        # Preserve any active option we don't normally display.
        for option_id in active:
            if option_id in shown:
                continue
            check = Gtk.CheckButton(label=self._option_description(group, option_id))
            check.set_active(True)
            check.connect("toggled", self._on_check_toggled, group, option_id)
            self.detail_box.pack_start(check, False, False, 0)

    def _build_single(self, group, active):
        active_id = active[0] if active else None

        none_radio = Gtk.RadioButton(label=_("None"))
        none_radio.set_active(active_id is None)
        none_radio.connect("toggled", self._on_radio_toggled, group, None)
        self.detail_box.pack_start(none_radio, False, False, 0)

        shown = set()
        for option_id, description in group["displayed"]:
            radio = Gtk.RadioButton.new_with_label_from_widget(none_radio, description)
            radio.set_active(option_id == active_id)
            radio.connect("toggled", self._on_radio_toggled, group, option_id)
            self.detail_box.pack_start(radio, False, False, 0)
            shown.add(option_id)

        # Preserve an active option we don't normally display.
        if active_id is not None and active_id not in shown:
            radio = Gtk.RadioButton.new_with_label_from_widget(
                none_radio, self._option_description(group, active_id))
            radio.set_active(True)
            radio.connect("toggled", self._on_radio_toggled, group, active_id)
            self.detail_box.pack_start(radio, False, False, 0)

    # --------------------------------------------------------------- handlers

    def _on_group_selected(self, listbox, row):
        if row is None:
            return
        self._current_group = row.group
        self._build_detail(row.group)

    def _on_radio_toggled(self, button, group, option_id):
        if self._building or not button.get_active():
            return
        self._write_group(group, [] if option_id is None else [option_id])

    def _on_check_toggled(self, button, group, option_id):
        if self._building:
            return
        active = set(self._stored_for_group(group))
        if button.get_active():
            active.add(option_id)
        else:
            active.discard(option_id)
        self._write_group(group, list(active))

    def _on_external_change(self, settings, key):
        if self._internal_write:
            return
        for group in self.groups:
            self._update_subtitle(group)
        self._resort()
        if self._current_group is not None:
            self._build_detail(self._current_group)

    # ------------------------------------------------------------ write/state

    def _write_group(self, group, new_group_options):
        current = self._stored_options()

        # Keep everything that doesn't belong to this group untouched .
        others = [o for o in current if o not in group["all_ids"]]
        new_list = others + new_group_options

        self._internal_write = True
        self.settings.set_strv("xkb-options", new_list)
        GLib.idle_add(self._clear_internal_write)

        self._update_subtitle(group)
        self._resort()

    def _clear_internal_write(self):
        self._internal_write = False
        return GLib.SOURCE_REMOVE

    def _update_subtitle(self, group):
        row = self._rows_by_group_id.get(group["id"])
        if row is None:
            return

        active = self._stored_for_group(group)
        if active:
            descriptions = [self._option_description(group, o) for o in active]
            # Multi-select groups can have several active options; list each on
            # its own line. Single-choice groups only ever have one.
            separator = "\n" if group["multi"] else ", "
            row.subtitle.set_text(separator.join(descriptions))
            row.subtitle.show()
        else:
            row.subtitle.hide()



# -*- coding=utf-8 -*-

import subprocess
import sys
import gettext
import json

gettext.install("cinnamon", "/usr/share/locale")

if __name__ == "__main__":
    results = []
    packages = subprocess.check_output(["apt-cache", "search"] + sys.argv[1].split(" ")).splitlines()[:10]
    for p in packages:
        i = p.index(" - ")
        name = p[:i]
        description = p[i+3:]
        results.append({'id': name, 'label': _("Install package : ") + name, 'description': description})

    print(json.dumps(results))



import argparse
import gettext
import sys

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk

import screenshot_backend
import prefs
import util

gettext.install('cinnamon', '/usr/share/locale')


class ScreenshotApplication:
    def __init__(self, args):
        self.args = args
        self.backend = screenshot_backend.Backend()
        self._exit_code = 0

    def quit(self):
        Gtk.main_quit()

    def run(self):
        self._activate()
        Gtk.main()
        return self._exit_code

    def _activate(self):
        args = self.args
        if args.clipboard:
            self._run_clipboard()
        elif args.file and not args.interactive:
            self._run_save_to_file(args.file)
        elif args.interactive:
            self._run_window()
        elif prefs.get_autosave_to_file():
            filename = util.build_filename(
                prefs.get_save_directory(),
                file_type=prefs.get_default_file_type(),
            )
            self._run_save_to_file(filename, copy_to_clipboard=prefs.get_autosave_to_clipboard())
        else:
            self._run_capture_to_editor()

    def _resolve_mode(self):
        if self.args.select_window:
            return 'select_window'
        if self.args.window:
            return 'window'
        if self.args.area:
            return 'area'
        if self.args.monitor is not None:
            return 'monitor'
        return 'screen'

    def capture(self, mode, include_pointer, include_shadow, delay, on_done, area_rect=None):
        # Monitor mode and area mode both end up calling screenshot_area,
        # which both grabs and flashes the captured rect on the server side.
        # Area mode without a rect runs the selector first to get one;
        # select_window mode runs the window picker first for an id.
        window_id = None

        def do_capture():
            if area_rect is not None:
                x, y, w, h = area_rect
                self.backend.screenshot_area(x, y, w, h, include_pointer, on_done)
            elif window_id is not None:
                self.backend.screenshot_window_by_id(window_id, include_pointer, include_shadow, on_done)
            elif mode == 'window':
                self.backend.screenshot_window(include_pointer, include_shadow, on_done)
            else:
                self.backend.screenshot(include_pointer, on_done)
            return GLib.SOURCE_REMOVE

        def schedule(fn):
            if delay > 0:
                GLib.timeout_add_seconds(delay, fn)
            else:
                GLib.idle_add(fn)

        if mode == 'area' and area_rect is None:
            def after_select(rect):
                nonlocal area_rect
                if rect is None:
                    on_done(None)
                    return
                area_rect = rect
                schedule(do_capture)
            self.backend.select_area(after_select)
        elif mode == 'select_window':
            def after_pick(wid):
                nonlocal window_id
                if wid is None:
                    on_done(None)
                    return
                window_id = wid
                schedule(do_capture)
            self.backend.select_window(after_pick)
        else:
            schedule(do_capture)

    def _run_window(self):
        from main_window import MainWindow
        win = MainWindow(self)
        win.run()

    def _run_capture_to_editor(self):
        # Non-interactive launch (e.g. the screenshot keybinding): grab first
        # and only build the editor window once we have an image to show.
        mode = self._resolve_mode()
        area_rect = None
        if mode == 'monitor':
            area_rect = util.monitor_rect(self.args.monitor)
            if area_rect is None:
                mode = 'screen'
        include_pointer = self.args.include_pointer
        include_shadow = self.args.include_shadow
        delay = self.args.delay if self.args.delay is not None else 0

        def done(pixbuf):
            from main_window import MainWindow
            win = MainWindow(self)
            win.show_with_pixbuf(pixbuf)

        self.capture(mode, include_pointer, include_shadow, delay, done, area_rect=area_rect)

    def _run_clipboard(self):
        mode = self._resolve_mode()
        area_rect = None
        if mode == 'monitor':
            area_rect = util.monitor_rect(self.args.monitor)
            if area_rect is None:
                mode = 'screen'
        include_pointer = self.args.include_pointer or prefs.get_include_pointer()
        include_shadow = self.args.include_shadow or prefs.get_include_shadow()
        delay = self.args.delay if self.args.delay is not None else 0

        def done(pixbuf):
            if pixbuf is not None:
                util.copy_pixbuf_to_clipboard(pixbuf)
                GLib.idle_add(self._finish_clipboard)
            else:
                self._exit_code = 1
                self.quit()

        self.capture(mode, include_pointer, include_shadow, delay, done, area_rect=area_rect)

    def _finish_clipboard(self):
        self.quit()
        return GLib.SOURCE_REMOVE

    def _run_save_to_file(self, path, copy_to_clipboard=False):
        mode = self._resolve_mode()
        area_rect = None
        if mode == 'monitor':
            area_rect = util.monitor_rect(self.args.monitor)
            if area_rect is None:
                mode = 'screen'
        include_pointer = self.args.include_pointer or prefs.get_include_pointer()
        include_shadow = self.args.include_shadow or prefs.get_include_shadow()
        delay = self.args.delay if self.args.delay is not None else 0

        def done(pixbuf):
            if pixbuf is not None:
                try:
                    util.save_pixbuf(pixbuf, path)
                    if copy_to_clipboard:
                        util.copy_pixbuf_to_clipboard(pixbuf)
                        GLib.idle_add(self._finish_clipboard)
                        return
                except Exception as exc:
                    print(f'cinnamon-screenshot: save failed: {exc}', file=sys.stderr)
                    self._exit_code = 1
            else:
                self._exit_code = 1
            self.quit()

        self.capture(mode, include_pointer, include_shadow, delay, done, area_rect=area_rect)


def _build_arg_parser():
    parser = argparse.ArgumentParser(
        prog='cinnamon-screenshot',
        description='Take screenshots of your screen, windows, or selected areas',
        add_help=True,
    )
    parser.add_argument('-c', '--clipboard', action='store_true',
                        help='Send the grab directly to the clipboard')
    parser.add_argument('-w', '--window', action='store_true',
                        help='Grab the active window instead of the entire screen')
    parser.add_argument('--select-window', action='store_true', dest='select_window',
                        help='Interactively pick a window to grab')
    parser.add_argument('-a', '--area', action='store_true',
                        help='Grab a selected area of the screen')
    parser.add_argument('-m', '--monitor', type=int, default=None, metavar='INDEX',
                        help='Grab a specific monitor by 0-based index')
    parser.add_argument('-p', '--include-pointer', action='store_true',
                        help='Include the pointer in the screenshot')
    parser.add_argument('-s', '--include-shadow', action='store_true',
                        help='Include the window shadow when grabbing a window')
    parser.add_argument('-d', '--delay', type=int, default=None, metavar='SECONDS',
                        help='Take the screenshot after a delay')
    parser.add_argument('-i', '--interactive', action='store_true',
                        help='Interactively set options before taking the screenshot')
    parser.add_argument('-f', '--file', metavar='PATH',
                        help='Save the screenshot directly to PATH')
    return parser

def main():
    parser = _build_arg_parser()
    args = parser.parse_args()

    target_flags = [args.window, args.select_window, args.area, args.monitor is not None]
    if sum(bool(f) for f in target_flags) > 1:
        parser.error('--window, --select-window, --area, and --monitor are mutually exclusive')
    if args.include_shadow and not (args.window or args.select_window):
        parser.error('--include-shadow requires --window or --select-window')
    if args.interactive and (args.delay is not None or args.include_pointer or args.include_shadow):
        parser.error('--interactive cannot be combined with --delay, --include-pointer, or --include-shadow')

    app = ScreenshotApplication(args)
    return app.run()



#!/usr/bin/python3

from gi.repository import Gtk, Gdk, GObject
from bin import util

print("KeybindingWidgets session type: %s" % util.get_session_type())

FORBIDDEN_UNMODDED_KEYVALS = [
    Gdk.KEY_Home,
    Gdk.KEY_Left,
    Gdk.KEY_Up,
    Gdk.KEY_Right,
    Gdk.KEY_Down,
    Gdk.KEY_Page_Up,
    Gdk.KEY_Page_Down,
    Gdk.KEY_End,
    Gdk.KEY_Tab,
    Gdk.KEY_Return,
    Gdk.KEY_space,
    Gdk.KEY_Mode_switch,
    Gdk.KEY_KP_0, # numerics currently are recognized only as _End, _Down, etc.. with or without numlock
    Gdk.KEY_KP_1, # Gdk checks numlock and parses out the correct key, but this could change, so list
    Gdk.KEY_KP_2, # these numerics anyhow. (This may differ depending on kb layouts, locales, etc.. but
    Gdk.KEY_KP_3, # I didn't thoroughly check.)
    Gdk.KEY_KP_4,
    Gdk.KEY_KP_5,
    Gdk.KEY_KP_6,
    Gdk.KEY_KP_7,
    Gdk.KEY_KP_8,
    Gdk.KEY_KP_9,
    Gdk.KEY_KP_End,
    Gdk.KEY_KP_Down,
    Gdk.KEY_KP_Next,
    Gdk.KEY_KP_Left,
    Gdk.KEY_KP_Begin,
    Gdk.KEY_KP_Right,
    Gdk.KEY_KP_Home,
    Gdk.KEY_KP_Up,
    Gdk.KEY_KP_Prior,
    Gdk.KEY_KP_Insert,
    Gdk.KEY_KP_Delete,
    Gdk.KEY_KP_Add,
    Gdk.KEY_KP_Subtract,
    Gdk.KEY_KP_Multiply,
    Gdk.KEY_KP_Divide,
    Gdk.KEY_KP_Enter,
    Gdk.KEY_Num_Lock
]

# Keys that can never be used in a shortcut at all (regardless of modifiers),
# paired with the reason shown to the user. (keyval, explanation)
FORBIDDEN_KEYVALS = [
    (Gdk.KEY_Caps_Lock,
     _("Caps Lock cannot be used in a keyboard shortcut - pressing it always "
       "toggles the Caps Lock state. You can configure this key to switch "
       "keyboard layouts in the XKB Options tab.")),
]

class ButtonKeybinding(Gtk.TreeView):
    __gsignals__ = {
        'accel-edited': (GObject.SignalFlags.RUN_LAST, None, (str, str)),
        'accel-cleared': (GObject.SignalFlags.RUN_LAST, None, ())
    }

    __gproperties__ = {
        "accel-string": (str,
                         "accelerator string",
                         "Parseable accelerator string",
                         None,
                         GObject.ParamFlags.READWRITE)
    }

    def __init__(self, position=0, **kwargs):
        super(ButtonKeybinding, self).__init__(**kwargs)
        self.position = position

        self.set_headers_visible(False)
        self.set_enable_search(False)
        self.set_hover_selection(True)
        self.set_tooltip_text(CellRendererKeybinding.TOOLTIP_TEXT)

        self.entry_store = None
        self.accel_string = ""
        self.keybinding_cell = CellRendererKeybinding(a_widget=self)
        self.keybinding_cell.set_alignment(.5,.5)
        self.keybinding_cell.connect('accel-edited', self.on_cell_edited)
        self.keybinding_cell.connect('accel-cleared', self.on_cell_cleared)

        col = Gtk.TreeViewColumn("binding", self.keybinding_cell, accel_string=0)
        col.set_alignment(.5)

        self.append_column(col)

        self.keybinding_cell.set_property('editable', True)

        self.load_model()

        self.connect("focus-out-event", self.on_focus_lost)

    def on_cell_edited(self, cell, path, accel_string, accel_label):
        self.accel_string = accel_string
        self.emit("accel-edited", accel_string, accel_label)
        self.load_model()

    def on_cell_cleared(self, cell, path):
        self.accel_string = ""
        self.emit("accel-cleared")
        self.load_model()

    def on_focus_lost(self, widget, event):
        self.get_selection().unselect_all()

    def load_model(self):
        if self.entry_store:
            self.entry_store.clear()

        self.entry_store = Gtk.ListStore(str) # Accel string
        self.entry_store.append((self.accel_string,))

        self.set_model(self.entry_store)

    def do_get_property(self, prop):
        if prop.name == 'accel-string':
            return self.accel_string
        else:
            raise AttributeError(f'unknown property {prop.name}')

    def do_set_property(self, prop, value):
        if prop.name == 'accel-string':
            if value != self.accel_string:
                self.accel_string = value
                self.keybinding_cell.set_value(value)
        else:
            raise AttributeError(f'unknown property {prop.name}')

    def get_accel_string(self):
        return self.accel_string

    def set_accel_string(self, accel_string):
        self.accel_string = accel_string
        self.load_model()


class CellRendererKeybinding(Gtk.CellRendererText):
    __gsignals__ = {
        'accel-edited': (GObject.SignalFlags.RUN_LAST, None, (str, str, str)),
        'accel-cleared': (GObject.SignalFlags.RUN_LAST, None, (str,))
    }

    __gproperties__ = {
        "accel-string": (str,
                         "accelerator string",
                         "Parseable accelerator string",
                         None,
                         GObject.ParamFlags.READWRITE)
    }

    # l10n constants - get them translated here so they use Cinnamon's gettext context
    TOOLTIP_TEXT = "%s\n%s\n%s" % (_("Click to set a new accelerator key."),
                                   _("Press Escape or click again to cancel the operation."),
                                   _("Press Backspace to clear the existing keybinding."))
    UNASSIGNED = _("unassigned")
    PICK_AN_ACCELERATOR = _("Pick an accelerator")
    MSG = _("\nThis key combination, \'<b>%s</b>\' should not be used because it would become impossible to type using this key. ")
    MSG += _("Please try again using a modifier key such as Control, Alt or Super (Windows key).\n\n")
    MSG += _("Continue only if you are certain this is what you want, otherwise press cancel.\n")
    CANCEL = _("Cancel")
    CONTINUE = _("Continue")

    def __init__(self, a_widget, accel_string=None):
        super(CellRendererKeybinding, self).__init__()
        self.connect("editing-started", self.editing_started)
        self.release_event_id = 0
        self.press_event_id = 0
        self.focus_id = 0

        self.a_widget = a_widget
        self.accel_string = accel_string

        self.path = None
        self.press_event = None
        self.teaching = False
        self.default_value = True
        self.text_string = ""
        self.seat = None
        self.keyboard = None

        self.update_label()

    def do_get_property(self, prop):
        if prop.name == 'accel-string':
            return self.accel_string
        else:
            raise AttributeError(f'unknown property {prop.name}')

    def do_set_property(self, prop, value):
        if prop.name == 'accel-string':
            if value != self.accel_string:
                self.accel_string = value
                self.update_label()
        else:
            raise AttributeError(f'unknown property {prop.name}')

    def update_label(self):
        text = CellRendererKeybinding.UNASSIGNED if self.default_value else self.text_string
        if self.accel_string:
            restore_atab = False
            restore_keyboard = False

            valid = self.accel_string
            if "Above_Tab" in valid:
                restore_atab = True
                valid = valid.replace("Above_Tab", "grave")
            # XF86Keyboard isn't recognized by accelerator_parse(), nor does Gdk.KEY_Keyboard
            # return a useful label by accelerator_get_label() (it returns the hex value of
            # that constant as a string - '0x1008ffb3').
            #
            # Parse some other key instead so we can at least get the modifiers right,
            # then restore 'Keyboard'.
            if "XF86Keyboard" in valid:
                restore_keyboard = True
                valid = valid.replace("XF86Keyboard", "End")
            key, codes, mods = Gtk.accelerator_parse_with_keycode(valid)
            if codes is not None and len(codes) > 0:
                text = Gtk.accelerator_get_label_with_keycode(None, key, codes[0], mods)
            if restore_atab:
                text = text.replace("`", "AboveTab")
            elif restore_keyboard:
                text = text.replace("End", "Keyboard")

        self.set_property("text", text)

    def set_value(self, accel_string=None):
        self.set_property("accel-string", accel_string)

    def editing_started(self, renderer, editable, path):
        if not self.teaching:
            self.path = path

            if util.get_session_type() == "x11":
                device = Gtk.get_current_event_device()
                if device.get_source() == Gdk.InputSource.KEYBOARD:
                    self.keyboard = device
                else:
                    self.keyboard = device.get_associated_device()

                self.keyboard.grab(self.a_widget.get_window(), Gdk.GrabOwnership.WINDOW, False,
                                   Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.KEY_RELEASE_MASK,
                                   None, Gdk.CURRENT_TIME)
            else:
                display = self.a_widget.get_display()
                self.seat = display.get_default_seat()

                # Grab both keyboard and pointer to prevent mouse events from canceling the operation
                grab_status = self.seat.grab(
                    self.a_widget.get_window(),
                    Gdk.SeatCapabilities.KEYBOARD | Gdk.SeatCapabilities.POINTER,
                    False,
                    None,
                    None,
                    None
                )

                if grab_status != Gdk.GrabStatus.SUCCESS:
                    print(f"Warning: Keyboard grab failed with status: {grab_status}")

            editable.set_text(CellRendererKeybinding.PICK_AN_ACCELERATOR)
            self.accel_editable = editable

            self.release_event_id = self.accel_editable.connect( "key-release-event", self.on_key_release )
            self.press_event_id = self.accel_editable.connect( "key-press-event", self.on_key_press )
            self.focus_id = self.accel_editable.connect( "focus-out-event", self.on_focus_out )
            self.teaching = True
        else:
            self.ungrab()
            self.update_label()
            self.teaching = False

    def on_focus_out(self, widget, event):
        self.teaching = False
        self.ungrab()

    def on_key_press(self, widget, event):
        if self.teaching:
            self.press_event = event.copy()
            return True

        return False

    def on_key_release(self, widget, event):
        self.ungrab()
        self.teaching = False

        if self.press_event is None:
            return True
        event = self.press_event

        display = widget.get_display()

        keyval = 0
        group = event.group
        accel_mods = event.state

        # HACK: we don't want to use SysRq as a keybinding (but we do
        # want Alt+Print), so we avoid translation from Alt+Print to SysRq

        if event.keyval == Gdk.KEY_Sys_Req and \
           ((accel_mods & Gdk.ModifierType.MOD1_MASK) != 0):
            keyval = Gdk.KEY_Print
            consumed_modifiers = 0
        else:
            keymap = Gdk.Keymap.get_for_display(display)
            group_mask_disabled = False
            shift_group_mask = 0

            shift_group_mask = keymap.get_modifier_mask(Gdk.ModifierIntent.SHIFT_GROUP)

            retval, keyval, effective_group, level, consumed_modifiers = \
                keymap.translate_keyboard_state(event.hardware_keycode, accel_mods, group)

            if group_mask_disabled:
                effective_group = 1

            if consumed_modifiers:
                consumed_modifiers &= ~shift_group_mask

        accel_key = Gdk.keyval_to_lower(keyval)
        if accel_key == Gdk.KEY_ISO_Left_Tab:
            accel_key = Gdk.KEY_Tab

        accel_mods &= Gtk.accelerator_get_default_mod_mask()

        # Some keys can never serve as a shortcut no matter the modifiers (e.g.
        # Caps Lock always toggles its own lock state). Reject them outright and
        # tell the user why.
        forbidden_msg = next((msg for kv, msg in FORBIDDEN_KEYVALS if kv == keyval), None)
        if forbidden_msg is not None:
            dialog = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                       Gtk.MessageType.WARNING,
                                       Gtk.ButtonsType.CLOSE,
                                       None)
            dialog.set_markup(forbidden_msg)
            dialog.show_all()
            dialog.run()
            dialog.destroy()

            self.update_label()
            self.teaching = False
            self.path = None
            self.press_event = None
            return True

        if accel_mods == 0:
            if accel_key == Gdk.KEY_Escape:
                self.update_label()
                self.teaching = False
                self.path = None
                self.press_event = None
                return True
            elif accel_key == Gdk.KEY_BackSpace:
                self.teaching = False
                self.press_event = None
                self.set_value(None)
                self.emit("accel-cleared", self.path)
                self.path = None
                return True

        accel_string = Gtk.accelerator_name_with_keycode(None, accel_key, event.hardware_keycode, Gdk.ModifierType(accel_mods))
        accel_label = Gtk.accelerator_get_label_with_keycode(None, accel_key, event.hardware_keycode, Gdk.ModifierType(accel_mods))

        # print("accel_mods: %d, keyval: %d, Storing %s as %s" % (accel_mods, keyval, accel_label, accel_string))

        if (accel_mods == 0 or accel_mods == Gdk.ModifierType.SHIFT_MASK) and event.hardware_keycode != 0 and self.default_value:
            if ((keyval >= Gdk.KEY_a                    and keyval <= Gdk.KEY_z)
                or  (keyval >= Gdk.KEY_A                    and keyval <= Gdk.KEY_Z)
                or  (keyval >= Gdk.KEY_0                    and keyval <= Gdk.KEY_9)
                or  (keyval >= Gdk.KEY_kana_fullstop        and keyval <= Gdk.KEY_semivoicedsound)
                or  (keyval >= Gdk.KEY_Arabic_comma         and keyval <= Gdk.KEY_Arabic_sukun)
                or  (keyval >= Gdk.KEY_Serbian_dje          and keyval <= Gdk.KEY_Cyrillic_HARDSIGN)
                or  (keyval >= Gdk.KEY_Greek_ALPHAaccent    and keyval <= Gdk.KEY_Greek_omega)
                or  (keyval >= Gdk.KEY_hebrew_doublelowline and keyval <= Gdk.KEY_hebrew_taf)
                or  (keyval >= Gdk.KEY_Thai_kokai           and keyval <= Gdk.KEY_Thai_lekkao)
                or  (keyval >= Gdk.KEY_Hangul               and keyval <= Gdk.KEY_Hangul_Special)
                or  (keyval >= Gdk.KEY_Hangul_Kiyeog        and keyval <= Gdk.KEY_Hangul_J_YeorinHieuh)
                    or  keyval in FORBIDDEN_UNMODDED_KEYVALS):
                dialog = Gtk.MessageDialog(None,
                                           Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                           Gtk.MessageType.WARNING,
                                           Gtk.ButtonsType.NONE,
                                           None)
                button = dialog.add_button(CellRendererKeybinding.CANCEL, Gtk.ResponseType.CANCEL)
                button = dialog.add_button(CellRendererKeybinding.CONTINUE, Gtk.ResponseType.OK)
                dialog.set_default_response(Gtk.ResponseType.CANCEL)
                button.get_style_context().add_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION)
                dialog.set_default_size(400, 200)
                dialog.set_markup(CellRendererKeybinding.MSG % accel_label)
                dialog.show_all()
                response = dialog.run()
                dialog.destroy()
                if response != Gtk.ResponseType.OK:
                    return True

        self.press_event = None
        self.set_value(accel_string)
        self.emit("accel-edited", self.path, accel_string, accel_label)
        self.path = None

        return True

    def ungrab(self):
        if util.get_session_type() == "x11":
            if self.keyboard is not None:
                self.keyboard.ungrab(Gdk.CURRENT_TIME)
                self.keyboard = None
        else:
            if self.seat is not None:
                self.seat.ungrab()
                self.seat = None
        if self.release_event_id > 0:
            self.accel_editable.disconnect(self.release_event_id)
            self.release_event_id = 0
        if self.press_event_id > 0:
            self.accel_editable.disconnect(self.press_event_id)
            self.press_event_id = 0
        if self.focus_id > 0:
            self.accel_editable.disconnect(self.focus_id)
            self.focus_id = 0
        try:
            self.accel_editable.editing_done()
            self.accel_editable.remove_widget()
        except:
            pass



#!/usr/bin/python3

import os
import gettext
import glob
import sys

from mintcommon import additionalfiles

import traceback

DOMAIN = "cinnamon"
PATH = "/usr/share/locale"

os.environ['LANGUAGE'] = "en_US.UTF-8"
gettext.install(DOMAIN, PATH)

try:
    sys.path.append('files/usr/share/cinnamon/cinnamon-settings')
    sys.path.append('files/usr/share/cinnamon/cinnamon-settings/modules')
    mod_files = glob.glob('files/usr/share/cinnamon/cinnamon-settings/modules/*.py')
    mod_files.sort()
    if len(mod_files) == 0:
        print("No settings modules found!!")
        sys.exit(1)

    mod_files = [x.split('/')[-1].split('.')[0] for x in mod_files]

    for mod_file in mod_files:
        if mod_file[0:3] != "cs_":
            raise Exception("Settings modules must have a prefix of 'cs_' !!")

    modules = map(__import__, mod_files)
except Exception as detail:
    print(detail)
    sys.exit(1)

for module in modules:
    try:
        mod = module.Module(None)
        name = mod.name

        if name == "display":
            # skip the display module, its desktop file is provided by CCC
            continue
        elif name == "accessibility":
            # rename modules which cs_*.py filename doesn't match their .desktop filename
            name = "universal-access"

        if mod.category in "admin":
            category = "Settings;System;"
        else:
            category = "Settings;"

        formatted_keywords = mod.sidePage.keywords.replace(",", ";")
        formatted_keywords = formatted_keywords.replace(", ", ";")

        prefix = """[Desktop Entry]
Icon=%(icon)s
Exec=cinnamon-settings %(module)s
Type=Application
OnlyShowIn=X-Cinnamon;
Categories=Settings;
""" % {'module': mod.name, 'category': category, 'icon': mod.sidePage.icon}

        additionalfiles.generate(DOMAIN, PATH, "files/usr/share/applications/cinnamon-settings-%s.desktop" % name, prefix, mod.sidePage.name, mod.comment, "", None, mod.sidePage.keywords)

    except Exception:
        print("Failed to load module %s" % module)
        traceback.print_exc()



#!/usr/bin/python3

from bin.SettingsWidgets import SidePage
from xapp.GSettingsWidgets import *

SCHEMA = "org.cinnamon"

class Module:
    name = "effects"
    category = "appear"
    comment = _("Control Cinnamon visual effects.")

    def __init__(self, content_box):
        keywords = _("effects, window")
        sidePage = SidePage(_("Effects"), "cs-desktop-effects", keywords, content_box, module=self)
        self.sidePage = sidePage

    def on_module_selected(self):
        if not self.loaded:
            print("Loading Effects module")

            self.schema = Gio.Settings(SCHEMA)

            page = SettingsPage()
            self.sidePage.add_widget(page)

            switch = GSettingsSwitch("", "org.cinnamon", "desktop-effects-workspace")
            switch.label.set_markup("<b>%s</b>" % _("Desktop and window effects"))
            switch.fill_row()
            page.add(switch)

            settings = page.add_reveal_section("", "org.cinnamon", "desktop-effects-workspace")

            widget = GSettingsSwitch(_("Effects on dialog boxes"), "org.cinnamon", "desktop-effects-on-dialogs")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects-workspace")

            widget = GSettingsSwitch(_("Effects on menus"), "org.cinnamon", "desktop-effects-on-menus")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects-workspace")

            widget = GSettingsSwitch(_("Session startup animation"), "org.cinnamon", "startup-animation")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects-workspace")

            widget = GSettingsSwitch(_("Fade effect on Cinnamon scrollboxes (like the Menu application list)"), "org.cinnamon", "enable-vfade")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects-workspace")


            widget = GSettingsSwitch(_("Window effects"), "org.cinnamon", "desktop-effects")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects-workspace")

            # MAPPING WINDOWS
            options = ["none", _("None")], \
                      ["traditional", _("Traditional")], \
                      ["fly", _("Fly")], \
                      ["move", _("Move")]
            widget = GSettingsComboBox(_("New windows or unminimizing existing ones"), "org.cinnamon", "desktop-effects-map", options)
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects")

            # CLOSING WINDOWS
            options = ["none", _("None")], \
                      ["traditional", _("Traditional")], \
                      ["fly", _("Fly")]
            widget = GSettingsComboBox(_("Closing windows"), "org.cinnamon", "desktop-effects-close", options)
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects")

            # MINIMIZING WINDOWS
            options = ["none", _("None")], \
                      ["traditional", _("Traditional")], \
                      ["fly", _("Fly")], \
                      ["fade", _("Fade")]
            widget = GSettingsComboBox(_("Minimizing windows"), "org.cinnamon", "desktop-effects-minimize", options)
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects")

            # MAXIMIZING/TILING WINDOWS
            widget = GSettingsSwitch(_("Resizing and tiling windows"), "org.cinnamon", "desktop-effects-change-size")
            settings.add_reveal_row(widget, "org.cinnamon", "desktop-effects")

            slider = GSettingsRange(_("Window animation speed"), "org.cinnamon", "window-effect-speed", _("Slower"), _("Faster"),
                                    mini=0, maxi=2, step=1, show_value=False)
            slider.content_widget.set_has_origin(False)
            slider.content_widget.add_mark(1, Gtk.PositionType.TOP, None)

            settings.add_reveal_row(slider, "org.cinnamon", "desktop-effects")

